from lxml import etree
from urllib.parse import urlparse
from celery import group
from celery.utils.log import get_logger

from ..import app
from . import detchar
from . import gcn
from . import gracedb
from . import external_skymaps
from . import igwn_alert
from . import raven

log = get_logger(__name__)


REQUIRED_LABELS_BY_TASK = {
    # EM_READY implies preferred event sky map is available
    'compare': {'EM_READY', 'EXT_SKYMAP_READY', 'EM_COINC'},
    'SoG': {'SKYMAP_READY', 'RAVEN_ALERT', 'ADVOK'}
}
"""These labels should be present on an external event to consider it to
be ready for sky map comparison or for post-alert analsis, such as a
measurment of the speed of gravity (SoG).
"""

FERMI_GRB_CLASS_VALUE = 4
"""This is the index that denote GRBs within Fermi's Flight Position
classification."""

FERMI_GRB_CLASS_THRESH = 50
"""This values denotes the threshold of the most likely Fermi source
classification, above which we will consider a Fermi Flight Position
notice."""


@gcn.handler(gcn.NoticeType.SNEWS,
             queue='exttrig',
             shared=False)
def handle_snews_gcn(payload):
    """Handles the payload from SNEWS alerts.

    Prepares the alert to be sent to graceDB as 'E' events.
    """
    root = etree.fromstring(payload)

    #  Get TrigID and Test Event Boolean
    trig_id = root.find("./What/Param[@name='TrigID']").attrib['value']
    ext_group = 'Test' if root.attrib['role'] == 'test' else 'External'

    event_observatory = 'SNEWS'
    if 'mdc-test_event' in root.attrib['ivorn'].lower():
        search = 'MDC'
    else:
        search = 'Supernova'
    query = 'group: External pipeline: {} grbevent.trigger_id = "{}"'.format(
        event_observatory, trig_id)
    events = gracedb.get_events(query=query)

    if events and ext_group == 'External':
        # Only update event if real
        assert len(events) == 1, 'Found more than one matching GraceDB entry'
        event, = events
        graceid = event['graceid']
        canvas = gracedb.replace_event.s(graceid, payload)

    else:
        canvas = gracedb.create_event.s(filecontents=payload,
                                        search=search,
                                        group=ext_group,
                                        pipeline=event_observatory)
        canvas |= _launch_external_detchar.s()

    canvas.delay()


@gcn.handler(gcn.NoticeType.FERMI_GBM_ALERT,
             gcn.NoticeType.FERMI_GBM_FLT_POS,
             gcn.NoticeType.FERMI_GBM_GND_POS,
             gcn.NoticeType.FERMI_GBM_FIN_POS,
             gcn.NoticeType.SWIFT_BAT_GRB_POS_ACK,
             gcn.NoticeType.FERMI_GBM_SUBTHRESH,
             gcn.NoticeType.INTEGRAL_WAKEUP,
             gcn.NoticeType.INTEGRAL_REFINED,
             gcn.NoticeType.INTEGRAL_OFFLINE,
             gcn.NoticeType.AGILE_MCAL_ALERT,
             queue='exttrig',
             shared=False)
def handle_grb_gcn(payload):
    """Handles the payload from Fermi, Swift, INTEGRAL, and AGILE MCAL
    GCN notices.

    Filters out candidates likely to be noise. Creates external events
    from the notice if new notice, otherwise updates existing event. Then
    creates and/or grabs external sky map to be uploaded to the external event.

    More info for these notices can be found at:
    Fermi-GBM: https://gcn.gsfc.nasa.gov/fermi_grbs.html
    Fermi-GBM sub: https://gcn.gsfc.nasa.gov/fermi_gbm_subthresh_archive.html
    Swift: https://gcn.gsfc.nasa.gov/swift.html
    INTEGRAL: https://gcn.gsfc.nasa.gov/integral.html
    AGILE-MCAL: https://gcn.gsfc.nasa.gov/agile_mcal.html
    """
    root = etree.fromstring(payload)
    u = urlparse(root.attrib['ivorn'])
    stream_path = u.path

    stream_obsv_dict = {'/SWIFT': 'Swift',
                        '/Fermi': 'Fermi',
                        '/INTEGRAL': 'INTEGRAL',
                        '/AGILE': 'AGILE'}
    event_observatory = stream_obsv_dict[stream_path]

    #  Get TrigID
    if event_observatory == 'INTEGRAL' and \
            not any([x in u.fragment for x in ['O3-replay', 'MDC-test']]):
        #  FIXME: revert all this if INTEGRAL fixes their GCN notices
        #  If INTEGRAL, get trigger ID from ivorn rather than the TrigID field
        #  unless O3 replay or MDC event
        trig_id = u.fragment.split('_')[-1].split('-')[0]
        #  Modify the TrigID field so GraceDB has the correct value
        root.find("./What/Param[@name='TrigID']").attrib['value'] = \
            str(trig_id).encode()
        #  Apply changes to payload delivered to GraceDB
        payload = etree.tostring(root, xml_declaration=True, encoding="UTF-8")
    else:
        try:
            trig_id = \
                root.find("./What/Param[@name='TrigID']").attrib['value']
        except AttributeError:
            trig_id = \
                root.find("./What/Param[@name='Trans_Num']").attrib['value']
    ext_group = 'Test' if root.attrib['role'] == 'test' else 'External'

    notice_type = \
        int(root.find("./What/Param[@name='Packet_Type']").attrib['value'])

    reliability = root.find("./What/Param[@name='Reliability']")
    if reliability is not None and int(reliability.attrib['value']) <= 4:
        return

    #  Check if Fermi trigger is likely noise by checking classification
    #  Most_Likely_Index of 4 is an astrophysical GRB
    #  If not at least 50% chance of GRB we will not consider it for RAVEN
    likely_source = root.find("./What/Param[@name='Most_Likely_Index']")
    likely_prob = root.find("./What/Param[@name='Most_Likely_Prob']")
    not_likely_grb = likely_source is not None and \
        (likely_source.attrib['value'] != FERMI_GRB_CLASS_VALUE
         or likely_prob.attrib['value'] < FERMI_GRB_CLASS_THRESH)

    #  Check if initial Fermi alert. These are generally unreliable and should
    #  never trigger a RAVEN alert, but will give us earlier warning of a
    #  possible coincidence. Later notices could change this.
    initial_gbm_alert = notice_type == gcn.NoticeType.FERMI_GBM_ALERT

    #  Check if Swift has lost lock. If so then veto
    lost_lock = \
        root.find("./What/Group[@name='Solution_Status']" +
                  "/Param[@name='StarTrack_Lost_Lock']")
    swift_veto = lost_lock is not None and lost_lock.attrib['value'] == 'true'

    #  Only send alerts if likely a GRB, is not a low-confidence early Fermi
    #  alert, and if not a Swift veto
    if not_likely_grb or initial_gbm_alert or swift_veto:
        labels = ['NOT_GRB']
    else:
        labels = None

    ivorn = root.attrib['ivorn']
    if 'subthresh' in ivorn.lower():
        search = 'SubGRB'
    elif 'mdc-test_event' in ivorn.lower():
        search = 'MDC'
    else:
        search = 'GRB'

    query = 'group: External pipeline: {} grbevent.trigger_id = "{}"'.format(
        event_observatory, trig_id)
    events = gracedb.get_events(query=query)

    group_canvas = ()
    if events and ext_group == 'External':
        # Only update event if real
        assert len(events) == 1, 'Found more than one matching GraceDB entry'
        event, = events
        graceid = event['graceid']
        if labels:
            canvas = gracedb.create_label.si(labels[0], graceid)
        else:
            canvas = gracedb.remove_label.si('NOT_GRB', graceid)

        # Prevent SubGRBs from appending GRBs
        if search == 'GRB':
            # Replace event and pass already existing event dictionary
            canvas |= gracedb.replace_event.si(graceid, payload)
            canvas |= gracedb.get_event.si(graceid)
        else:
            return

    else:
        canvas = gracedb.create_event.s(filecontents=payload,
                                        search=search,
                                        group=ext_group,
                                        pipeline=event_observatory,
                                        labels=labels)
        group_canvas += _launch_external_detchar.s(),

    if search in {'GRB', 'MDC'}:
        notice_date = root.find("./Who/Date").text
        group_canvas += external_skymaps.create_upload_external_skymap.s(
                      notice_type, notice_date),
    if event_observatory == 'Fermi':
        if search == 'SubGRB':
            skymap_link = \
                root.find("./What/Param[@name='HealPix_URL']").attrib['value']
            group_canvas += \
                external_skymaps.get_upload_external_skymap.s(skymap_link),
        elif search == 'GRB':
            skymap_link = None
            group_canvas += \
                external_skymaps.get_upload_external_skymap.s(skymap_link),

    (
        canvas
        |
        group(group_canvas)
    ).delay()


@igwn_alert.handler('superevent',
                    'mdc_superevent',
                    'external_fermi',
                    'external_swift',
                    'external_integral',
                    'external_agile',
                    queue='exttrig',
                    shared=False)
def handle_grb_igwn_alert(alert):
    """Parse an IGWN alert message related to superevents/GRB external triggers
    and dispatch it to other tasks.

    Notes
    -----
    This IGWN alert message handler is triggered by creating a new superevent
    or GRB external trigger event, or a label associated with completeness of
    skymaps:

    * Any new event triggers a coincidence search with
      :meth:`gwcelery.tasks.raven.coincidence_search`.
    * When both a GW and GRB sky map are available during a coincidence,
      indicated by the labels ``EM_READY`` and ``EXT_SKYMAP_READY``
      respectively on the external event, this triggers the spacetime coinc
      FAR to be calculated and a combined GW-GRB sky map is created using
      :meth:`gwcelery.tasks.external_skymaps.create_combined_skymap`.

    """
    # Determine GraceDB ID
    graceid = alert['uid']

    # launch searches
    if alert['alert_type'] == 'new':
        if alert['object'].get('group') == 'External':
            # Create and upload Swift sky map for the joint targeted
            # sub-threshold search as agreed on in the MOU
            if alert['object']['search'] == 'SubGRBTargeted' and \
                    alert['object']['pipeline'] == 'Swift':
                external_skymaps.create_upload_external_skymap(
                    alert['object'], None, alert['object']['created'])

            # launch search with MDC events and exit
            if alert['object']['search'] == 'MDC':
                raven.coincidence_search(graceid, alert['object'],
                                         group='CBC', se_searches=['MDC'])
                raven.coincidence_search(graceid, alert['object'],
                                         group='Burst', se_searches=['MDC'])
                return

            # launch standard Burst-GRB search
            raven.coincidence_search(graceid, alert['object'], group='Burst',
                                     se_searches=['Allsky'])

            if alert['object']['search'] in ['SubGRB', 'SubGRBTargeted']:
                # if sub-threshold GRB, launch search with that pipeline
                raven.coincidence_search(
                    graceid, alert['object'], group='CBC',
                    searches=['SubGRB', 'SubGRBTargeted'],
                    pipelines=[alert['object']['pipeline']])
            else:
                # if threshold GRB, launch standard CBC-GRB search
                raven.coincidence_search(graceid, alert['object'],
                                         group='CBC', searches=['GRB'])
        elif 'S' in graceid:
            # launch standard GRB search based on group
            gw_group = alert['object']['preferred_event_data']['group']

            # launch search with MDC events and exit
            if alert['object']['preferred_event_data']['search'] == 'MDC':
                raven.coincidence_search(graceid, alert['object'],
                                         group=gw_group, searches=['MDC'])
                return

            if gw_group == 'CBC':
                # launch subthreshold searches if CBC
                # for Fermi and Swift separately to use different time windows
                for pipeline in ['Fermi', 'Swift']:
                    raven.coincidence_search(
                        graceid, alert['object'], group='CBC',
                        searches=['SubGRB', 'SubGRBTargeted'],
                        pipelines=[pipeline])
                se_searches = []
            else:
                se_searches = ['Allsky']
            # launch standard GRB search
            raven.coincidence_search(graceid, alert['object'],
                                     group=gw_group, searches=['GRB'],
                                     se_searches=se_searches)
    # re-run raven pipeline and create combined sky map (if not a Swift event)
    # when sky maps are available
    elif alert['alert_type'] == 'label_added' and \
            alert['object'].get('group') == 'External':
        if _skymaps_are_ready(alert['object'], alert['data']['name'],
                              'compare'):
            # if both sky maps present and a coincidence, re-run RAVEN
            # pipeline and create combined sky maps
            ext_event = alert['object']
            superevent_id, ext_id = _get_superevent_ext_ids(
                graceid, ext_event)
            superevent = gracedb.get_superevent(superevent_id)
            _relaunch_raven_pipeline_with_skymaps(
                superevent, ext_event, graceid)
        elif 'EM_COINC' in alert['object']['labels']:
            # if not complete, check if GW sky map; apply label to external
            # event if GW sky map
            se_labels = gracedb.get_labels(alert['object']['superevent'])
            if 'SKYMAP_READY' in se_labels:
                gracedb.create_label.si('SKYMAP_READY', graceid).delay()
            if 'EM_READY' in se_labels:
                gracedb.create_label.si('EM_READY', graceid).delay()
    # apply labels from superevent to external event to update state
    # and trigger functionality requiring sky maps, etc.
    elif alert['alert_type'] == 'label_added' and 'S' in graceid:
        if 'SKYMAP_READY' in alert['object']['labels']:
            # if sky map in superevent, apply label to all external events
            # at the time
            group(
                gracedb.create_label.si('SKYMAP_READY', ext_id)
                for ext_id in alert['object']['em_events']
            ).delay()
        if 'EM_READY' in alert['object']['labels']:
            # if sky map not in superevent but in preferred event, apply label
            # to all external events at the time
            group(
                gracedb.create_label.si('EM_READY', ext_id)
                for ext_id in alert['object']['em_events']
            ).delay()
        if _skymaps_are_ready(alert['object'], alert['data']['name'], 'SoG') \
                and alert['object']['space_coinc_far'] is not None:
            # if a superevent is vetted by ADVOK and a spatial joint FAR is
            # available, check if SoG publishing conditions are met
            (
                gracedb.get_event.si(alert['object']['em_type'])
                |
                raven.sog_paper_pipeline.s(alert['object'])
            ).delay()
    # if new GW or external sky map after first being available, try to remake
    # combine sky map and rerun raven pipeline
    elif alert['alert_type'] == 'log' and \
            'EM_COINC' in alert['object']['labels'] and \
            'fit' in alert['data']['filename'] and \
            'flat' not in alert['data']['comment'].lower() and \
            (alert['data']['filename'] not in
             {external_skymaps.COMBINED_SKYMAP_FILENAME_MULTIORDER,
              external_skymaps.COMBINED_SKYMAP_FILENAME_FLAT}):
        superevent_id, external_id = _get_superevent_ext_ids(
                                         graceid, alert['object'])
        if 'S' in graceid:
            superevent = alert['object']
        else:
            superevent = gracedb.get_superevent(alert['object']['superevent'])
            external_event = alert['object']
        # check if combined sky map already made, with the exception of Swift
        # which will fail
        if 'S' in graceid:
            # Rerun for all eligible external events
            for ext_id in superevent['em_events']:
                external_event = gracedb.get_event(ext_id)
                if REQUIRED_LABELS_BY_TASK['compare'].issubset(
                        set(external_event['labels'])):
                    _relaunch_raven_pipeline_with_skymaps(
                        superevent, external_event, graceid,
                        use_superevent=True)
        else:
            if REQUIRED_LABELS_BY_TASK['compare'].issubset(
                    set(external_event['labels'])):
                _relaunch_raven_pipeline_with_skymaps(
                    superevent, external_event, graceid)
    elif alert['alert_type'] == 'label_removed' and \
            alert['object'].get('group') == 'External':
        if alert['data']['name'] == 'NOT_GRB' and \
                'EM_COINC' in alert['object']['labels']:
            # if NOT_GRB is removed, re-check publishing conditions
            superevent_id = alert['object']['superevent']
            superevent = gracedb.get_superevent(superevent_id)
            gw_group = superevent['preferred_event_data']['group']
            coinc_far_dict = {
                'temporal_coinc_far': superevent['time_coinc_far'],
                'spatiotemporal_coinc_far': superevent['space_coinc_far']
            }
            raven.trigger_raven_alert(coinc_far_dict, superevent, graceid,
                                      alert['object'], gw_group)


@igwn_alert.handler('superevent',
                    'mdc_superevent',
                    'external_snews',
                    queue='exttrig',
                    shared=False)
def handle_snews_igwn_alert(alert):
    """Parse an IGWN alert message related to superevents/SN external triggers
    and dispatch it to other tasks.

    Notes
    -----
    This igwn_alert message handler is triggered by creating a new superevent
    or SN external trigger event:

    * Any new event triggers a coincidence search with
      :meth:`gwcelery.tasks.raven.coincidence_search`.

    """
    # Determine GraceDB ID
    graceid = alert['uid']

    if alert['alert_type'] == 'new':
        if alert['object'].get('superevent_id'):
            group = alert['object']['preferred_event_data']['group']
            search = alert['object']['preferred_event_data']['search']
            if search == 'MDC':
                raven.coincidence_search(graceid, alert['object'],
                                         group='Burst', searches=['MDC'],
                                         pipelines=['SNEWS'])
            # Run on Test and Burst superevents
            elif group in {'Burst', 'Test'}:
                raven.coincidence_search(graceid, alert['object'],
                                         group='Burst', searches=['Supernova'],
                                         pipelines=['SNEWS'])
        else:
            # Run on SNEWS event, either real or test
            search = alert['object']['search']
            if search == 'MDC':
                raven.coincidence_search(graceid, alert['object'],
                                         group='Burst',
                                         se_searches=['MDC'],
                                         pipelines=['SNEWS'])
            elif search == 'Supernova':
                raven.coincidence_search(graceid, alert['object'],
                                         group='Burst', searches=['Supernova'],
                                         pipelines=['SNEWS'])


def _skymaps_are_ready(event, label, task):
    label_set = set(event['labels'])
    required_labels = REQUIRED_LABELS_BY_TASK[task]
    return required_labels.issubset(label_set) and label in required_labels


def _get_superevent_ext_ids(graceid, event):
    if 'S' in graceid:
        se_id = event['superevent_id']
        ext_id = event['em_type']
    else:
        se_id = event['superevent']
        ext_id = event['graceid']
    return se_id, ext_id


@app.task(queue='exttrig',
          shared=False)
def _launch_external_detchar(event):
    start = event['gpstime']
    if event['pipeline'] == 'SNEWS':
        start, end = event['gpstime'], event['gpstime']
    else:
        integration_time = \
            event['extra_attributes']['GRB']['trigger_duration'] or 4.0
        end = start + integration_time
    detchar.check_vectors.si(event, event['graceid'], start, end).delay()

    return event


def _relaunch_raven_pipeline_with_skymaps(superevent, ext_event, graceid,
                                          use_superevent=False):
    """Relaunch the RAVEN sky map comparison workflow, include recalculating
    the joint FAR with updated sky map info and create a new combined sky map.

    Parameters
    ----------
    superevent: dict
        superevent dictionary
    exttrig: dict
        external event dictionary
    graceid: str
        GraceDB ID of event
    use_superevent: bool
        If True, always use skymap info from superevent
        regardless of SKYMAP_READY label.

    """
    gw_group = superevent['preferred_event_data']['group']
    tl, th = raven._time_window(graceid, gw_group,
                                [ext_event['pipeline']],
                                [ext_event['search']])
    # FIXME: both overlap integral and combined sky map could be
    # done by the same function since they are so similar
    canvas = raven.raven_pipeline.si(
                 [ext_event] if 'S' in graceid else [superevent],
                 graceid,
                 superevent if 'S' in graceid else ext_event,
                 tl, th, gw_group)
    # Swift localizations are incredibly well localized and require
    # a different method from Fermi/Integral/AGILE
    # FIXME: Add Swift localization information in the future
    if ext_event['pipeline'] != 'Swift':
        # Create new updated combined sky map
        canvas |= external_skymaps.create_combined_skymap.si(
                      superevent['superevent_id'], ext_event['graceid'],
                      preferred_event=(
                          None if 'SKYMAP_READY' in ext_event['labels']
                          or use_superevent
                          else superevent['preferred_event']))
    canvas.delay()



"""Application configuration for ``gracedb.ligo.org``.

Inherits all settings from :mod:`gwcelery.conf.playground`, with the exceptions
below.
"""

from base64 import b64encode

from . import *  # noqa: F401, F403

condor_accounting_group = 'ligo.prod.o4.cbc.pe.bayestar'
"""HTCondor accounting group for Celery workers launched with condor_submit."""

expose_to_public = True
"""Set to True if events meeting the public alert threshold really should be
exposed to the public."""

igwn_alert_group = 'gracedb'
"""IGWN alert group."""

gracedb_host = 'gracedb.ligo.org'
"""GraceDB host."""

create_mattermost_channel = True
"""Create Mattermost channel in production"""

kafka_alert_config = {
    'scimma': {'url': 'kafka://kafka.scimma.org/igwn.gwalert',
               'suffix': 'avro', 'skymap_encoder': lambda _: _},
    'gcn': {'url': 'kafka://kafka.gcn.nasa.gov/igwn.gwalert',
            'suffix': 'json', 'skymap_encoder': lambda b:
            b64encode(b).decode('utf-8')}
}
"""Kafka broker configuration details"""

# FIXME kafka_consumer_config is only empty here as a safety precaution for
# ER15 production deployment. We need to deploy a new release because we need
# to enable alerts, but we haven't tested the Kafka consumer bootsteps enough
# to deploy them on production yet.
kafka_consumer_config = {}
"""Kafka consumer configuration details. The keys describe the senders of the
messages to be consumed. The values are a dictionary of the URL to listen to
and information about the message serializer."""

voevent_broadcaster_address = ':5341'
"""The VOEvent broker will bind to this address to send GCNs.
This should be a string of the form `host:port`. If `host` is empty,
then listen on all available interfaces."""

voevent_broadcaster_whitelist = ['capella2.gsfc.nasa.gov']
"""List of hosts from which the broker will accept connections.
If empty, then completely disable the broker's broadcast capability."""

llhoft_glob = '/dev/shm/kafka/{detector}/*.gwf'
"""File glob for low-latency h(t) frames."""

low_latency_frame_types = {'H1': 'H1_llhoft',
                           'L1': 'L1_llhoft',
                           'V1': 'V1_llhoft'}
"""Types of frames used in Parameter Estimation (see
:mod:`gwcelery.tasks.inference`) and in cache creation for detchar
checks (see :mod:`gwcelery.tasks.detchar`).
"""

high_latency_frame_types = {'H1': 'H1_HOFT_C00',
                            'L1': 'L1_HOFT_C00',
                            'V1': 'V1Online'}
"""Types of high latency frames used in Parameter Estimation
(see :mod:`gwcelery.tasks.inference`) and in cache creation for detchar
checks (see :mod:`gwcelery.tasks.detchar`).
"""

idq_channels = ['H1:IDQ-FAP_OVL_10_2048',
                'L1:IDQ-FAP_OVL_10_2048']
"""Low-latency iDQ false alarm probability channel names from live O3 frames"""

strain_channel_names = {'H1': 'H1:GDS-CALIB_STRAIN_CLEAN',
                        'L1': 'L1:GDS-CALIB_STRAIN_CLEAN',
                        'V1': 'V1:Hrec_hoft_16384Hz'}
"""Names of h(t) channels used in Parameter Estimation (see
:mod:`gwcelery.tasks.inference`)"""

sentry_environment = 'production'
"""Record this `environment tag
<https://docs.sentry.io/enriching-error-data/environments/>`_ in Sentry log
messages."""

only_alert_for_mdc = False
"""If True, then only sends alerts for MDC events. Useful for times outside
of observing runs."""

condor_retry_kwargs = dict(
    max_retries=None, retry_backoff=True, retry_jitter=True,
    retry_backoff_max=600
)
"""Retry settings of condor.submit task."""



from threading import Thread

from celery import bootsteps
from celery.utils.log import get_logger
from safe_netrc import netrc

from .signals import email_received

__all__ = ('Receiver',)

log = get_logger(__name__)


class EmailBootStep(bootsteps.ConsumerStep):
    """Generic boot step to limit us to appropriate kinds of workers.

    Only include this bootstep in workers that are started with the
    ``--email`` command line option.
    """

    def __init__(self, consumer, email=False, **kwargs):
        self.enabled = bool(email)

    def start(self, consumer):
        log.info('Starting %s', self.name)

    def stop(self, consumer):
        log.info('Stopping %s', self.name)


class Receiver(EmailBootStep):
    """Run the global email receiver in background thread."""

    name = 'email client'

    def _runloop(self):
        from imapclient import IMAPClient
        from imapclient.exceptions import IMAPClientAbortError

        username, _, password = netrc().authenticators(self._host)
        while self._running:
            try:
                log.debug('Starting new connection')
                with IMAPClient(self._host, use_uid=True, timeout=30) as conn:
                    log.debug('Logging in')
                    conn.login(username, password)
                    log.debug('Selecting inbox')
                    conn.select_folder('inbox')
                    while self._running:
                        log.debug('Searching for new messages')
                        messages = conn.search()
                        log.debug('Fetching new messages')
                        for msgid, data in conn.fetch(
                                messages, ['RFC822']).items():
                            log.debug('Sending signal for new email')
                            email_received.send(None, rfc822=data[b'RFC822'])
                            log.debug('Deleting email')
                            conn.delete_messages(msgid)
                        log.debug('Starting idle')
                        conn.idle()
                        # Stay in IDLE mode for at most 5 minutes.
                        # According to the imapclient documentation:
                        #
                        # > Note that IMAPClient does not handle low-level
                        # > socket errors that can happen when maintaining
                        # > long-lived TCP connections. Users are advised to
                        # > renew the IDLE command every 10 minutes to avoid
                        # > the connection from being abruptly closed.
                        for _ in range(60):
                            if not self._running or conn.idle_check(timeout=5):
                                break
                        log.debug('Idle done')
                        conn.idle_done()
            except IMAPClientAbortError:
                log.exception('IMAP connection aborted')

    def start(self, consumer):
        super().start(consumer)
        self._host = consumer.app.conf['email_host']
        self._running = True
        self._thread = Thread(target=self._runloop, name='EmailClientThread')
        self._thread.start()

    def stop(self, consumer):
        super().stop(consumer)
        self._running = False
        self._thread.join()



"""Generate and upload automated circulars."""
import ligo.followup_advocate

from . import legacy_gracedb as gracedb


@gracedb.task(shared=False)
def create_initial_circular(graceid):
    """Create and return circular txt."""
    return ligo.followup_advocate.compose(graceid, client=gracedb.client)


@gracedb.task(shared=False)
def create_emcoinc_circular(graceid):
    """Create and return the em_coinc circular txt."""
    return ligo.followup_advocate.compose_raven(graceid,
                                                client=gracedb.client)


@gracedb.task(shared=False)
def create_update_circular(graceid, update_types=['sky_localization',
                                                  'em_bright',
                                                  'p_astro']):
    """Create and return update circular txt."""
    return ligo.followup_advocate.compose_update(graceid,
                                                 update_types=update_types,
                                                 client=gracedb.client)


@gracedb.task(shared=False)
def create_retraction_circular(graceid):
    """Create and return retraction circular txt."""
    return ligo.followup_advocate.compose_retraction(graceid,
                                                     client=gracedb.client)



"""Definitions of custom :doc:`Celery signals <celery:userguide/signals>`
related to emails.

These signals allow us to keep the VOEvent validation code decoupled from the
email client itself.
"""
from celery.utils.dispatch import Signal

email_received = Signal(
    name='email_received', providing_args=('rfc822',))
"""Fired whenever an email message is received.

Parameters
----------
rfc822 : bytes
    The :rfc:`822` contents of the message.

Examples
--------

Register an email listener like this::

    import email
    import email.policy

    @email_received.connect
    def on_email_received(rfc822, **kwargs):
        # Parse the RFC822 email.
        message = email.message_from_bytes(rfc822, policy=email.policy.default)
        # Print some of the message headers.
        print('Subject:', message['Subject'])
        print('From:', message['From'])
        # Print the plain-text message body.
        body = message.get_body(['plain']).get_content()
        print(body)
"""



"""Error telemetry for `Sentry <https://sentry.io>`_."""
from urllib.parse import urlparse, urlunparse

from celery.utils.log import get_logger
from safe_netrc import netrc, NetrcParseError
import sentry_sdk
from sentry_sdk.integrations import celery, flask, logging, redis, tornado

from .. import _version
from ..util import SPHINX
from .integrations import condor, requests, subprocess

log = get_logger(__name__)

__all__ = ('configure', 'DSN')

DSN = 'https://sentry.io/1425216'
"""Sentry data source name (DSN)."""


def configure():
    """Configure Sentry logging integration for Celery.

    See the `official instructions for Celery integration
    <https://docs.sentry.io/platforms/python/celery/>`_.

    Notes
    -----
    Add the API key username/pasword pair to your netrc file.

    """
    # Catching NetrcParseError confuses sphinx.
    if SPHINX:  # pragma: no cover
        return

    # Delayed import
    from .. import app

    scheme, netloc, *rest = urlparse(DSN)

    try:
        auth = netrc().authenticators(netloc)
        if not auth:
            raise ValueError('No netrc entry found for {}'.format(netloc))
    except (NetrcParseError, OSError, ValueError):
        log.exception('Disabling Sentry integration because we could not load '
                      'the username and password for %s from the netrc file',
                      netloc)
        return

    # Ignore noisy log but harmless messsages from adc-streaming.
    #
    # FIXME: Remove once https://git.ligo.org/emfollow/gwcelery/-/issues/457
    # is fixed.
    logging.ignore_logger('adc-streaming')

    # The "legacy" Sentry DSN requires a "public key" and a "private key",
    # which are transmitted as the username and password in the URL.
    # However, as of Sentry 9, then "private key" part is no longer required.
    username, _, _ = auth
    dsn = urlunparse(
        (scheme, '{}@{}'.format(username, netloc), *rest))
    version = 'gwcelery-{}'.format(_version.get_versions()['version'])
    environment = app.conf['sentry_environment']
    sentry_sdk.init(dsn, environment=environment, release=version,
                    integrations=[celery.CeleryIntegration(),
                                  condor.CondorIntegration(),
                                  flask.FlaskIntegration(),
                                  redis.RedisIntegration(),
                                  requests.RequestsIntegration(),
                                  subprocess.SubprocessIntegration(),
                                  tornado.TornadoIntegration()])



"""Module containing the functionality for creation and management of
superevents.

*   There is serial processing of triggers from low latency pipelines.
*   Dedicated **superevent** queue for this purpose.
*   Primary logic to respond to low latency triggers contained in
    :meth:`process` function.
"""
from itertools import filterfalse, groupby

from celery.utils.log import get_task_logger
from ligo.segments import segment, segmentlist

from ..import app
from . import gracedb, igwn_alert

log = get_task_logger(__name__)

REQUIRED_LABELS_BY_GROUP = {
    'cbc': {'PASTRO_READY', 'EMBRIGHT_READY', 'SKYMAP_READY'},
    'burst': {'SKYMAP_READY'}
}
"""These labels should be present on an event to consider it to
be complete.
"""

FROZEN_LABEL = 'LOW_SIGNIF_LOCKED'
"""This label indicates that the superevent manager should make no further
changes to the preferred event."""

SIGNIFICANT_LABEL = 'SIGNIF_LOCKED'
"""This label indicates that the superevent is elevated to significant"""

EARLY_WARNING_LABEL = 'EARLY_WARNING'
"""This label indicates that the superevent contains a significant
early warning event."""

EARLY_WARNING_SEARCH_NAME = 'EarlyWarning'
"""Search name for Early Warning searches. Only significant events
result in
consideration by the superevent manager."""

READY_LABEL = 'EM_READY'
"""This label indicates that a preferred event has been assigned and it
has all data products required to make it ready for annotations."""


@igwn_alert.handler('cbc_gstlal',
                    'cbc_spiir',
                    'cbc_pycbc',
                    'cbc_mbta',
                    'burst_olib',
                    'burst_cwb',
                    'burst_mly',
                    shared=False)
def handle(payload):
    """Respond to IGWN alert topics from low-latency search pipelines and
    delegate to :meth:`process` for superevent management.
    """
    alert_type = payload['alert_type']
    gid = payload['object']['graceid']

    try:
        far = payload['object']['far']
    except KeyError:
        log.info('Skipping %s because it lacks a FAR', gid)
        return

    if far > app.conf['superevent_far_threshold']:
        log.info("Skipping processing of %s because of high FAR", gid)
        return

    search = payload['object']['search']
    if search == EARLY_WARNING_SEARCH_NAME and \
            far > app.conf['early_warning_alert_far_threshold'] / \
            app.conf['early_warning_alert_trials_factor']:
        log.info(
            "Skipping processing EW event %s because it does not"
            "meet significant event criterion", gid
        )
        return

    priority = 1
    if alert_type == 'label_added':
        priority = 0
        label = payload['data']['name']
        group = payload['object']['group'].lower()
        if label == 'RAVEN_ALERT':
            log.info('Label %s added to %s', label, gid)
        elif label not in REQUIRED_LABELS_BY_GROUP[group]:
            return
        elif not is_complete(payload['object']):
            log.info("Ignoring since %s has %s labels. "
                     "It is not complete", gid, payload['object']['labels'])
            return
    elif alert_type != 'new':
        return

    process.si(payload).apply_async(priority=priority)


@gracedb.task(queue='superevent', shared=False)
@gracedb.catch_retryable_http_errors
def process(payload):
    """
    Respond to `payload` and serially processes them to create new superevents,
    add events to existing ones.

    Parameters
    ----------
    payload : dict
        IGWN alert payload

    """
    event_info = payload['object']
    gid = event_info['graceid']
    category = get_category(event_info)
    t_0, t_start, t_end = get_ts(event_info)

    if event_info.get('superevent'):
        sid = event_info['superevent']
        log.info('Event %s already belongs to superevent %s', gid, sid)
        # superevent_neighbours has current and nearby superevents
        s = event_info['superevent_neighbours'][sid]
        superevent = _SuperEvent(s['t_start'],
                                 s['t_end'],
                                 s['t_0'],
                                 s['superevent_id'])
        _update_superevent(superevent.superevent_id,
                           event_info,
                           t_0=t_0,
                           t_start=None,
                           t_end=None)
    else:
        log.info('Event %s does not yet belong to a superevent', gid)
        superevents = gracedb.get_superevents('category: {} {} .. {}'.format(
            category,
            event_info['gpstime'] - app.conf['superevent_query_d_t_start'],
            event_info['gpstime'] + app.conf['superevent_query_d_t_end']))
        for s in superevents:
            if gid in s['gw_events']:
                sid = s['superevent_id']
                log.info('Event %s found assigned to superevent %s', gid, sid)
                if payload['alert_type'] == 'label_added':
                    log.info('Label %s added to %s',
                             payload['data']['name'], gid)
                elif payload['alert_type'] == 'new':
                    log.info('new alert type for %s with '
                             'existing superevent %s. '
                             'No action required', gid, sid)
                    return
                superevent = _SuperEvent(s['t_start'],
                                         s['t_end'],
                                         s['t_0'],
                                         s['superevent_id'])
                _update_superevent(superevent.superevent_id,
                                   event_info,
                                   t_0=t_0,
                                   t_start=None,
                                   t_end=None)
                break
        else:  # s not in superevents
            event_segment = _Event(t_start, t_end, t_0, event_info['graceid'])

            superevent = _partially_intersects(superevents,
                                               event_segment)

            if superevent:
                sid = superevent.superevent_id
                log.info('Event %s in window of %s. '
                         'Adding event to superevent', gid, sid)
                gracedb.add_event_to_superevent(sid, event_segment.gid)
                # extend the time window of the superevent
                new_superevent = superevent | event_segment
                if new_superevent != superevent:
                    log.info('%s not completely contained in %s, '
                             'extending superevent window',
                             event_segment.gid, sid)
                    new_t_start, new_t_end = new_superevent

                else:  # new_superevent == superevent
                    log.info('%s is completely contained in %s',
                             event_segment.gid, sid)
                    new_t_start = new_t_end = None
                _update_superevent(superevent.superevent_id,
                                   event_info,
                                   t_0=t_0,
                                   t_start=new_t_start,
                                   t_end=new_t_end)
            else:  # not superevent
                log.info('New event %s with no superevent in GraceDB, '
                         'creating new superevent', gid)
                sid = gracedb.create_superevent(event_info['graceid'],
                                                t_0, t_start, t_end)

    if should_publish(event_info, significant=True):
        gracedb.create_label.delay('ADVREQ', sid)
        if is_complete(event_info):
            if EARLY_WARNING_LABEL in event_info['labels']:
                gracedb.create_label(EARLY_WARNING_LABEL, sid)
            else:
                gracedb.create_label(SIGNIFICANT_LABEL, sid)

    if should_publish(event_info, significant=False):
        if is_complete(event_info):
            gracedb.create_label(FROZEN_LABEL, sid)


def get_category(event):
    """Get the superevent category for an event.

    Parameters
    ----------
    event : dict
        Event dictionary (e.g., the return value from
        :meth:`gwcelery.tasks.gracedb.get_event`).

    Returns
    -------
    {'mdc', 'test', 'production'}

    """
    if event.get('search') == 'MDC':
        return 'mdc'
    elif event['group'] == 'Test':
        return 'test'
    else:
        return 'production'


def get_ts(event):
    """Get time extent of an event, depending on pipeline-specific parameters.

    *   For CWB and MLy, use the event's ``duration`` field.
    *   For oLIB, use the ratio of the event's ``quality_mean`` and
        ``frequency_mean`` fields.
    *   For all other pipelines, use the
        :obj:`~gwcelery.conf.superevent_d_t_start` and
        :obj:`~gwcelery.conf.superevent_d_t_end` configuration values.

    Parameters
    ----------
    event : dict
        Event dictionary (e.g., the return value from
        :meth:`gwcelery.tasks.gracedb.get_event` or
        ``preferred_event_data`` in igwn-alert packet.)

    Returns
    -------
    t_0: float
        Segment center time in GPS seconds.
    t_start : float
        Segment start time in GPS seconds.

    t_end : float
        Segment end time in GPS seconds.

    """
    pipeline = event['pipeline'].lower()
    if pipeline == 'cwb':
        attribs = event['extra_attributes']['MultiBurst']
        d_t_start = d_t_end = attribs['duration']
    elif pipeline == 'olib':
        attribs = event['extra_attributes']['LalInferenceBurst']
        d_t_start = d_t_end = (attribs['quality_mean'] /
                               attribs['frequency_mean'])
    elif pipeline == 'mly':
        attribs = event['extra_attributes']['MLyBurst']
        d_t_start = d_t_end = attribs['duration']
    else:
        d_t_start = app.conf['superevent_d_t_start'].get(
            pipeline, app.conf['superevent_default_d_t_start'])
        d_t_end = app.conf['superevent_d_t_end'].get(
            pipeline, app.conf['superevent_default_d_t_end'])
    return (event['gpstime'], event['gpstime'] - d_t_start,
            event['gpstime'] + d_t_end)


def get_snr(event):
    """Get the SNR from the LVAlert packet.

    Different groups and pipelines store the SNR in different fields.

    Parameters
    ----------
    event : dict
        Event dictionary (e.g., the return value from
        :meth:`gwcelery.tasks.gracedb.get_event`, or
        ``preferred_event_data`` in igwn-alert packet.)

    Returns
    -------
    snr : float
        The SNR.

    """
    group = event['group'].lower()
    pipeline = event['pipeline'].lower()
    if group == 'cbc':
        attribs = event['extra_attributes']['CoincInspiral']
        return attribs['snr']
    elif pipeline == 'cwb':
        attribs = event['extra_attributes']['MultiBurst']
        return attribs['snr']
    elif pipeline == 'olib':
        attribs = event['extra_attributes']['LalInferenceBurst']
        return attribs['omicron_snr_network']
    elif pipeline == 'mly':
        attribs = event['extra_attributes']['MLyBurst']
        return attribs['SNR']
    else:
        raise NotImplementedError('SNR attribute not found')


def get_instruments(event):
    """Get the instruments that contributed data to an event.

    Parameters
    ----------
    event : dict
        Event dictionary (e.g., the return value from
        :meth:`gwcelery.tasks.gracedb.get_event`, or
        ``preferred_event_data`` in igwn-alert packet.)

    Returns
    -------
    set
        The set of instruments that contributed to the event.

    """
    attribs = event['extra_attributes']['SingleInspiral']
    ifos = {single['ifo'] for single in attribs}
    return ifos


def get_instruments_in_ranking_statistic(event):
    """Get the instruments that contribute to the false alarm rate.

    Parameters
    ----------
    event : dict
        Event dictionary (e.g., the return value from
        :meth:`gwcelery.tasks.gracedb.get_event`, or
        ``preferred_event_data`` in igwn-alert packet.)

    Returns
    -------
    set
        The set of instruments that contributed to the ranking statistic for
        the event.

    Notes
    -----
    The number of instruments that contributed *data* to an event is given by
    the ``instruments`` key of the GraceDB event JSON structure. However, some
    pipelines (e.g. gstlal) have a distinction between which instruments
    contributed *data* and which were considered in the *ranking* of the
    candidate. For such pipelines, we infer which pipelines contributed to the
    ranking by counting only the SingleInspiral records for which the chi
    squared field is non-empty.

    For PyCBC Live in the O3 configuration, an empty chi^2 field does not mean
    that the detector did not contribute to the ranking; in fact, *all*
    detectors listed in the SingleInspiral table contribute to the significance
    even if the chi^2 is not computed for some of them. Hence PyCBC Live is
    handled as a special case.

    """
    if event['pipeline'].lower() == 'pycbc':
        return set(event['instruments'].split(','))
    else:
        attribs = event['extra_attributes']['SingleInspiral']
        return {single['ifo'] for single in attribs
                if single.get('chisq') is not None}


@app.task(shared=False)
def select_pipeline_preferred_event(events):
    """Group list of G events by pipeline, and apply :meth:`keyfunc`
    to select the pipeline preferred events.

    Parameters
    ----------
    events : list
        list of event dictionaries

    Returns
    -------
    dict
        pipeline, graceid pairs
    """
    g_events = list(
        filterfalse(lambda x: x['group'] == "External", events))
    g_events_by_pipeline = groupby(
        sorted(g_events, key=lambda x: x['pipeline']),
        key=lambda x: x['pipeline']
    )

    return dict(
        (k, select_preferred_event(g)) for k, g in g_events_by_pipeline)


@app.task(shared=False)
def select_preferred_event(events):
    """Select the preferred event out of a list of G events, typically
    contents of a superevent, based on :meth:`keyfunc`.

    Parameters
    ----------
    events : list
        list of event dictionaries

    """
    g_events = list(
        filterfalse(lambda x: x['group'] == "External", events))
    return max(g_events, key=keyfunc)


def is_complete(event):
    """
    Determine if a G event is complete in the sense of the event
    has its data products complete i.e. has PASTRO_READY, SKYMAP_READY,
    EMBRIGHT_READY for CBC events and the SKYMAP_READY label for the
    Burst events. Test events are not processed by low-latency infrastructure
    and are always labeled complete.

    Parameters
    ----------
    event : dict
        Event dictionary (e.g., the return value from
        :meth:`gwcelery.tasks.gracedb.get_event`, or
        ``preferred_event_data`` in igwn-alert packet.)

    """
    group = event['group'].lower()
    label_set = set(event['labels'])
    required_labels = REQUIRED_LABELS_BY_GROUP[group]
    return required_labels.issubset(label_set)


def should_publish(event, significant=True):
    """Determine whether an event should be published as a public alert.

    All of the following conditions must be true for a public alert:

    *   The event's ``offline`` flag is not set.
    *   The event's false alarm rate, weighted by the group-specific trials
        factor as specified by the
        :obj:`~gwcelery.conf.preliminary_alert_trials_factor`
        (:obj:`~gwcelery.conf.significant_alert_trials_factor`)
        configuration setting, is less than or equal to
        :obj:`~gwcelery.conf.preliminary_alert_far_threshold`
        (:obj:`~gwcelery.conf.significant_alert_far_threshold`)

    Parameters
    ----------
    event : dict
        Event dictionary (e.g., the return value from
        :meth:`gwcelery.tasks.gracedb.get_event`, or
        ``preferred_event_data`` in igwn-alert packet.)
    significant : bool
        Flag to use significant
        (:obj:`~gwcelery.conf.significant_alert_far_threshold`),
        or less-significant
        (:obj:`~gwcelery.conf.preliminary_alert_far_threshold`)
        FAR threshold.

    Returns
    -------
    should_publish : bool
        :obj:`True` if the event meets the criteria for a public alert or
        :obj:`False` if it does not.

    """
    return all(_should_publish(event, significant=significant))


def _should_publish(event, significant=False):
    """Wrapper around :meth:`should_publish`. Returns the boolean returns of
    the publishability criteria as a tuple for later use.
    """
    group = event['group'].lower()
    if not significant:
        far_threshold = app.conf['preliminary_alert_far_threshold'][group]
        trials_factor = app.conf['preliminary_alert_trials_factor'][group]
    elif EARLY_WARNING_LABEL in event['labels']:
        far_threshold = app.conf['early_warning_alert_far_threshold']
        trials_factor = app.conf['early_warning_alert_trials_factor']
    else:
        far_threshold = app.conf['significant_alert_far_threshold'][group]
        trials_factor = app.conf['significant_alert_trials_factor'][group]
    far = trials_factor * event['far']
    raven_coincidence = ('RAVEN_ALERT' in event['labels'])

    return (not event['offline'] and 'INJ' not in event['labels'],
            far <= far_threshold or raven_coincidence)


def keyfunc(event):
    """Key function for selection of the preferred event.

    Return a value suitable for identifying the preferred event. Given events
    ``a`` and ``b``, ``a`` is preferred over ``b`` if
    ``keyfunc(a) > keyfunc(b)``, else ``b`` is preferred.

    Parameters
    ----------
    event : dict
        Event dictionary (e.g., the return value from
        :meth:`gwcelery.tasks.gracedb.get_event`).

    Returns
    -------
    key : tuple
        The comparison key.

    Notes
    -----
    Tuples are compared lexicographically in Python: they are compared
    element-wise until an unequal pair of elements is found.

    """
    group = event['group'].lower()
    try:
        group_rank = ['burst', 'cbc'].index(group)
    except ValueError:
        group_rank = -1

    if group == 'cbc':
        group_rank = 1
        n_ifos = len(get_instruments(event))
        snr_or_far_ranking = get_snr(event)
    else:
        # We don't care about the number of detectors for burst events.
        n_ifos = -1
        # Smaller FAR -> higher IFAR -> more significant.
        # Use -FAR instead of IFAR=1/FAR so that rank for FAR=0 is defined.
        snr_or_far_ranking = -event['far']

    # Conditions that determine choice of the preferred event
    # event completeness comes first
    # then, publishability criteria for significant events
    # then, publishability criteria for less-significant events
    # then, CBC group is given higher priority over Burst
    # then, prioritize more number of detectors
    # finally, use SNR (FAR) between two CBC (Burst) events
    # See https://rtd.igwn.org/projects/gwcelery/en/latest/gwcelery.tasks.superevents.html#selection-of-the-preferred-event  # noqa: E501
    return (
        is_complete(event),
        *_should_publish(event, significant=True),
        *_should_publish(event, significant=False),
        group_rank,
        n_ifos,
        snr_or_far_ranking
    )


def _update_superevent(superevent_id, new_event_dict,
                       t_0, t_start, t_end):
    """Update preferred event and/or change time window. Events with multiple
    detectors take precedence over single-detector events, then CBC events take
    precedence over burst events, and any remaining tie is broken by SNR/FAR
    values for CBC/Burst. Single detector are not promoted to preferred event
    status, if existing preferred event is multi-detector

    Parameters
    ----------
    superevent_id : str
        the superevent_id
    new_event_dict : dict
        event info of the new trigger as a dictionary
    t_0 : float
        center time of `superevent_id`, None for no change
    t_start : float
        start time of `superevent_id`, None for no change
    t_end : float
        end time of `superevent_id`, None for no change

    """
    # labels and preferred event in the IGWN alert are not the latest
    superevent_dict = gracedb.get_superevent(superevent_id)

    superevent_labels = superevent_dict['labels']
    preferred_event_dict = superevent_dict['preferred_event_data']
    kwargs = {}
    if t_start is not None:
        kwargs['t_start'] = t_start
    if t_end is not None:
        kwargs['t_end'] = t_end
    if FROZEN_LABEL not in superevent_labels:
        if keyfunc(new_event_dict) > keyfunc(preferred_event_dict):
            # update preferred event when LOW_SIGNIF_LOCKED is not applied
            kwargs['t_0'] = t_0
            kwargs['preferred_event'] = new_event_dict['graceid']

    if kwargs:
        gracedb.update_superevent(superevent_id, **kwargs)

    # completeness takes first precedence in deciding preferred event
    # necessary and suffiecient condition to superevent as ready
    if is_complete(new_event_dict):
        gracedb.create_label.delay(READY_LABEL, superevent_id)


def _superevent_segment_list(superevents):
    """Ingests a list of superevent dictionaries, and returns a segmentlist
    with start and end times as the duration of each segment.

    Parameters
    ----------
    superevents : list
        List of superevent dictionaries (e.g., the values
        of field ``superevent_neighbours`` in igwn-alert packet).

    Returns
    -------
    superevent_list : segmentlist
        superevents as a segmentlist object

    """
    return segmentlist(
        [_SuperEvent(s['t_start'], s['t_end'], s['t_0'], s['superevent_id'])
         for s in superevents])


def _partially_intersects(superevents, event_segment):
    """Similar to :meth:`segmentlist.find` except it also returns the segment
    of `superevents` which partially intersects argument. If there are more
    than one intersections, first occurence is returned.

    Parameters
    ----------
    superevents : list
        list of superevents. Typical value of
        ``superevent_neighbours.values()``.
    event_segment : segment
        segment object whose index is wanted

    Returns
    -------
    match_segment   : segment
        segment in `self` which intersects. `None` if not found

    """
    # create a segmentlist using start and end times
    superevents = _superevent_segment_list(superevents)
    for s in superevents:
        if s.intersects(event_segment):
            return s
    return None


class _Event(segment):
    """An event implemented as an extension of :class:`segment`."""

    def __new__(cls, t_start, t_end, *args, **kwargs):
        return super().__new__(cls, t_start, t_end)

    def __init__(self, t_start, t_end, t_0, gid):
        self.t_0 = t_0
        self.gid = gid


class _SuperEvent(segment):
    """An superevent implemented as an extension of :class:`segment`."""

    def __new__(cls, t_start, t_end, *args, **kwargs):
        return super().__new__(cls, t_start, t_end)

    def __init__(self, t_start, t_end, t_0, sid):
        self.t_start = t_start
        self.t_end = t_end
        self.t_0 = t_0
        self.superevent_id = sid



"""Matplotlib environment management."""
from contextlib import contextmanager

from matplotlib import pyplot as plt

__all__ = ('closing_figures',)


@contextmanager
def closing_figures():
    """Close figure that are created in a with: statement."""
    old_fignums = set(plt.get_fignums())
    try:
        yield
    finally:
        new_fignums = set(plt.get_fignums())
        for fignum in new_fignums - old_fignums:
            plt.close(fignum)



"""Data quality and detector characterization tasks.

These tasks are mostly focused on checking interferometer state vectors. By
design, the [LIGO]_ and [Virgo]_ state vectors share the same definitions for
the first 8 fields.

LIGO also has a [DMT]_ DQ vector that provides some additional instrumental
checks.

References
----------
.. [LIGO] https://wiki.ligo.org/Calibration/TDCalibReview
.. [Virgo] https://dcc.ligo.org/G1801125/
.. [DMT] https://wiki.ligo.org/DetChar/DmtDqVector

"""
import getpass
import glob
import io
import json
import socket
import time

from astropy.time import Time
from celery import group
from celery.utils.log import get_task_logger
from glue.lal import Cache
from gwdatafind import find_urls
from gwpy.timeseries import Bits, StateVector, TimeSeries
from gwpy.plot import Plot
import matplotlib.pyplot as plt
import numpy as np

from . import gracedb
from ..import app
from ..import _version
from ..jinja import env
from ..util import closing_figures


__author__ = 'Geoffrey Mo <geoffrey.mo@ligo.org>'

log = get_task_logger(__name__)


def create_cache(ifo, start, end):
    """Find .gwf files and create cache. Will first look in the llhoft, and
    if the frames have expired from llhoft, will call gwdatafind.

    Parameters
    ----------
    ifo : str
        Interferometer name (e.g. ``H1``).
    start, end: int or float
        GPS start and end times desired.

    Returns
    -------
    :class:`glue.lal.Cache`

    Example
    -------
    >>> create_cache('H1', 1198800018, 1198800618)
    [<glue.lal.CacheEntry at 0x7fbae6b71278>,
      <glue.lal.CacheEntry at 0x7fbae6ae5b38>,
      <glue.lal.CacheEntry at 0x7fbae6ae5c50>,
     ...
      <glue.lal.CacheEntry at 0x7fbae6b15080>,
      <glue.lal.CacheEntry at 0x7fbae6b15828>]

    """
    pattern = app.conf['llhoft_glob'].format(detector=ifo)
    filenames = glob.glob(pattern)
    cache = Cache.from_urls(filenames)

    try:
        cache_starttime = int(
            list(cache.to_segmentlistdict().values())[0][0][0])
    except IndexError:
        log.exception('Files do not exist in llhoft_glob')
        return cache  # returns empty cache

    if start >= cache_starttime:  # required data is in llhoft
        return cache

    # otherwise, required data has left llhoft
    high_latency = app.conf['high_latency_frame_types'][ifo]
    urls = find_urls(ifo[0], high_latency, start, end)
    if urls:
        return Cache.from_urls(urls)

    # required data not in high latency frames
    low_latency = app.conf['low_latency_frame_types'][ifo]
    urls = find_urls(ifo[0], low_latency, start, end)
    if not urls:  # required data not in low latency frames
        log.error('This data cannot be found, or does not exist.')

    return Cache.from_urls(urls)


@app.task(shared=False)
@closing_figures()
def make_omegascan(ifo, t0, durs):
    """Helper function to create a single omegascan image, with
    multiple durations.

    Parameters
    ----------
    ifo : str
        'H1', 'L1', or 'V1'
    t0 : int or float
        Central time of the omegascan.
    durs : list of tuples
        List of three tuples, with time before and time after t0 in seconds.
        Example: [(0.75, 0.25), (1.5, 0.5), (7.5, 2.5)]

    Returns
    -------
    bytes or None
        bytes of png of the omegascan, or None if no omegascan created.

    """
    # Explicitly use a non-interactive Matplotlib backend.
    plt.switch_backend('agg')

    # Collect data
    durs = np.array(durs)
    longest_before, longest_after = durs[:, 0].max(), durs[:, 1].max()
    # longest = max(np.array(durs).flatten())
    long_start, long_end = t0 - longest_before, t0 + longest_after
    cache = create_cache(ifo, long_start, long_end)
    strain_name = app.conf['strain_channel_names'][ifo]
    try:
        ts = TimeSeries.read(cache, strain_name,
                             start=long_start, end=long_end).astype('float64')
        # Do q_transforms for the different durations
        qgrams = [ts.q_transform(
            frange=(20, 4096), gps=t0,
            outseg=(t0 - before, t0 + after), logf=True)
            for before, after in durs]
    except (IndexError, FloatingPointError, ValueError):
        # data from cache can't be properly read, or data is weird
        fig = plt.figure()
        plt.axis("off")
        plt.text(0.1, 0.45, f"Failed to create {ifo} omegascan", fontsize=17)
    else:
        fig = Plot(*qgrams,
                   figsize=(12 * len(durs), 6),
                   geometry=(1, len(durs)),
                   yscale='log',
                   method='pcolormesh',
                   cmap='viridis')
        for i, ax in enumerate(fig.axes):
            if i in [1, 2]:
                ax.set_ylabel('')
            if i == 2:
                fig.colorbar(ax=ax, label='Normalized energy', clim=(0, 30))
            ax.set_epoch(t0)
        fig.suptitle(f'Omegascans of {strain_name} at {t0}', fontweight="bold")
    plt.subplots_adjust(wspace=0.08)
    outfile = io.BytesIO()
    fig.savefig(outfile, format='png', dpi=150, bbox_inches='tight')
    return outfile.getvalue()


@app.task(shared=False)
def omegascan(t0, graceid):
    """Create omegascan for a certain event.

    Parameters
    ----------
    t0 : float
        Central event time.
    graceid : str
        GraceDB ID to which to upload the omegascan.

    """
    durs = app.conf['omegascan_durations']
    durs = np.array(durs)

    # Delay for early warning events (ie queries for times before now)
    longest_after = durs[:, 1].max()
    if t0 + longest_after > Time.now().gps:
        log.info("Delaying omegascan because %s is in the future",
                 graceid)
        waittime = t0 - Time.now().gps + longest_after + 10
    else:
        waittime = longest_after + 10

    group(
        make_omegascan.s(ifo, t0, durs).set(countdown=waittime)
        |
        gracedb.upload.s(f"{ifo}_omegascan.png", graceid,
                         f"{ifo} omegascan", tags=['data_quality'])
        for ifo in ['H1', 'L1', 'V1']
    ).delay()


def generate_table(title, high_bit_list, low_bit_list, unknown_bit_list):
    """Make a nice table which shows the status of the bits checked.

    Parameters
    ----------
    title : str
        Title of the table.
    high_bit_list: list
        List of bit names which are high.
    low_bit_list: list
        List of bit names which are low.
    unknown_bit_list: list
        List of bit names which are unknown.

    Returns
    -------
    str
        HTML string of the table.

    """
    template = env.get_template('vector_table.jinja2')
    return template.render(title=title, high_bit_list=high_bit_list,
                           low_bit_list=low_bit_list,
                           unknown_bit_list=unknown_bit_list)


def dqr_json(state, summary):
    """Generate DQR-compatible json-ready dictionary from process results, as
    described in :class:`data-quality-report.design`.

    Parameters
    ----------
    state : {'pass', 'fail'}
        State of the detchar checks.
    summary : str
        Summary of results from the process.

    Returns
    -------
    dict
        Ready to be converted into json.

    """
    links = [
        {
            "href":
            "https://gwcelery.readthedocs.io/en/latest/gwcelery.tasks.detchar.html#gwcelery.tasks.detchar.check_vectors", # noqa
            "innerHTML": "a link to the documentation for this process"
        },
        {
            "href":
            "https://git.ligo.org/emfollow/gwcelery/blob/main/gwcelery/tasks/detchar.py",  # noqa
            "innerHTML": "a link to the source code in the gwcelery repo"
        }
    ]
    return dict(
        state=state,
        process_name=__name__,
        process_version=_version.get_versions()['version'],
        librarian='Geoffrey Mo (geoffrey.mo@ligo.org)',
        date=time.strftime("%H:%M:%S UTC %a %d %b %Y", time.gmtime()),
        hostname=socket.gethostname(),
        username=getpass.getuser(),
        summary=summary,
        figures=[],
        tables=[],
        links=links,
        extra=[],
    )


def check_idq(cache, channel, start, end):
    """Looks for iDQ frame and reads them.

    Parameters
    ----------
    cache : :class:`glue.lal.Cache`
        Cache from which to check.
    channel : str
        which idq channel (FAP)
    start, end: int or float
        GPS start and end times desired.

    Returns
    -------
    tuple
        Tuple mapping iDQ channel to its minimum FAP.

    Example
    -------
    >>> check_idq(cache, 'H1:IDQ-FAP_OVL_100_1000',
                  1216496260, 1216496262)
    ('H1:IDQ-FAP_OVL_100_1000', 0.003)

    """
    if cache:
        try:
            idq_fap = TimeSeries.read(
                cache, channel, start=start, end=end)
            return (channel, float(idq_fap.min().value))
        except (IndexError, RuntimeError):
            # FIXME: Change from log.exception to log.warning until this fixed,
            # because it's saturating Sentry.
            log.warning('Failed to read from low-latency iDQ frame files')
    # FIXME: figure out how to get access to low-latency frames outside
    # of the cluster. Until we figure that out, actual I/O errors have
    # to be non-fatal.
    return (channel, None)


def check_vector(cache, channel, start, end, bits, logic_type='all'):
    """Check timeseries of decimals against a bitmask.
    This is inclusive of the start time and exclusive of the end time, i.e.
    [start, ..., end).

    Parameters
    ----------
    cache : :class:`glue.lal.Cache`
        Cache from which to check.
    channel : str
        Channel to look at, e.g. ``H1:DMT-DQ_VECTOR``.
    start, end : int or float
        GPS start and end times desired.
    bits: :class:`gwpy.TimeSeries.Bits`
        Definitions of the bits in the channel.
    logic_type : str, optional
        Type of logic to apply for vetoing.
        If ``all``, then all samples in the window must pass the bitmask.
        If ``any``, then one or more samples in the window must pass.

    Returns
    -------
    dict
        Maps each bit in channel to its state.

    Example
    -------
    >>> check_vector(cache, 'H1:GDS-CALIB_STATE_VECTOR', 1216496260,
                     1216496262, ligo_state_vector_bits)
    {'H1:HOFT_OK': True,
     'H1:OBSERVATION_INTENT': True,
     'H1:NO_STOCH_HW_INJ': True,
     'H1:NO_CBC_HW_INJ': True,
     'H1:NO_BURST_HW_INJ': True,
     'H1:NO_DETCHAR_HW_INJ': True}

    """
    if logic_type not in ('any', 'all'):
        raise ValueError("logic_type must be either 'all' or 'any'.")
    else:
        logic_map = {'any': np.any, 'all': np.all}
    bitname = '{}:{}'
    if cache:
        try:
            statevector = StateVector.read(cache, channel,
                                           start=start, end=end, bits=bits)
        except (IndexError, TypeError):
            # FIXME: TypeError above is due to
            # https://github.com/gwpy/gwpy/issues/1211
            #
            # FIXME: Change from log.exception to log.warning until this fixed,
            # because it's saturating Sentry.
            log.warning('Failed to read from low-latency frame files')
        else:
            # FIXME: In the playground environment, the Virgo state vector
            # channel is stored as a float. Is this also the case in the
            # production environment?
            statevector = statevector.astype(np.uint32)
            if len(statevector) > 0:  # statevector must not be empty
                return {bitname.format(channel.split(':')[0], key):
                        bool(logic_map[logic_type](
                            value.value if len(value.value) > 0 else None))
                        for key, value in statevector.get_bit_series().items()}
    # FIXME: figure out how to get access to low-latency frames outside
    # of the cluster. Until we figure that out, actual I/O errors have
    # to be non-fatal.
    return {bitname.format(channel.split(':')[0], key):
            None for key in bits if key is not None}


@app.task(shared=False)
def check_vectors(event, graceid, start, end):
    """Perform data quality checks for an event and labels/logs results to
    GraceDB.

    Depending on the pipeline, a certain amount of time (specified in
    :obj:`~gwcelery.conf.check_vector_prepost`) is appended to either side of
    the superevent start and end time. This is to catch DQ issues slightly
    before and after the event, such as that appearing in L1 just before
    GW170817.

    A cache is then created for H1, L1, and V1, regardless of the detectors
    involved in the event. Then, the bits and channels specified in the
    configuration file (:obj:`~gwcelery.conf.llhoft_channels`) are checked.
    If an injection is found in the active detectors, 'INJ' is labeled to
    GraceDB. If an injection is found in any detector, a message with the
    injection found is logged to GraceDB. If no injections are found across
    all detectors, this is logged to GraceDB.

    A similar task is performed for the DQ states described in the
    DMT-DQ_VECTOR, LIGO GDS-CALIB_STATE_VECTOR, and Virgo
    DQ_ANALYSIS_STATE_VECTOR. If no DQ issues are found in active detectors,
    'DQOK' is labeled to GraceDB. Otherwise, 'DQV' is labeled. In all cases,
    the DQ states of all the state vectors checked are logged to GraceDB.

    This skips MDC events.

    Parameters
    ----------
    event : dict
        Details of event.
    graceid : str
        GraceID of event to which to log.
    start, end : int or float
        GPS start and end times desired.

    Returns
    -------
    event : dict
        Details of the event, reflecting any labels that were added.

    """
    # Skip early warning events (ie queries for times before now)
    if end > Time.now().gps:
        log.info("Skipping detchar checks because %s is in the future",
                 event['graceid'])
        return event

    # Skip MDC events.
    if event.get('search') == 'MDC':
        log.info("Skipping detchar checks because %s is an MDC",
                 event['graceid'])
        return event

    # Create caches for all detectors
    instruments = event['instruments'].split(',')
    pipeline = event['pipeline']
    pre, post = app.conf['check_vector_prepost'][pipeline]
    start, end = start - pre, end + post
    prepost_msg = "Check looked within -{}/+{} seconds of superevent. ".format(
        pre, post)

    ifos = {key.split(':')[0] for key, val in
            app.conf['llhoft_channels'].items()}
    caches = {ifo: create_cache(ifo, start, end) for ifo in ifos}
    bit_defs = {channel_type: Bits(channel=bitdef['channel'],
                                   bits=bitdef['bits'])
                for channel_type, bitdef
                in app.conf['detchar_bit_definitions'].items()}

    # Examine injection and DQ states
    # Do not analyze DMT-DQ_VECTOR if pipeline uses gated h(t)
    states = {}
    analysis_channels = app.conf['llhoft_channels'].items()
    if app.conf['uses_gatedhoft'][pipeline]:
        analysis_channels = {k: v for k, v in analysis_channels
                             if k[3:] != 'DMT-DQ_VECTOR'}.items()
    for channel, bits in analysis_channels:
        states.update(check_vector(caches[channel.split(':')[0]], channel,
                                   start, end, bit_defs[bits]))
    # Pick out DQ and injection states, then filter for active detectors
    dq_states = {key: value for key, value in states.items()
                 if key.split('_')[-1] != 'INJ'}
    inj_states = {key: value for key, value in states.items()
                  if key.split('_')[-1] == 'INJ'}
    active_dq_states = {key: value for key, value in dq_states.items()
                        if key.split(':')[0] in instruments}
    active_inj_states = {key: value for key, value in inj_states.items()
                         if key.split(':')[0] in instruments}

    # Check iDQ states
    idq_faps = dict(check_idq(caches[channel.split(':')[0]],
                    channel, start, end)
                    for channel in app.conf['idq_channels'])

    # Logging iDQ to GraceDB
    if None not in idq_faps.values():
        idq_faps_readable = {k: round(v, 3) for k, v in idq_faps.items()}
        if min(idq_faps.values()) <= app.conf['idq_fap_thresh']:
            idq_msg = ("iDQ false alarm probability is low: "
                       "minimum FAP is {}. ").format(
                json.dumps(idq_faps_readable)[1:-1])
            # If iDQ FAP is low and pipeline enabled, apply DQV
            if app.conf['idq_veto'][pipeline]:
                gracedb.remove_label('DQOK', graceid)
                gracedb.create_label('DQV', graceid)
                # Add labels to return value to avoid querying GraceDB again.
                event = dict(event, labels=event.get('labels', []) + ['DQV'])
                try:
                    event['labels'].remove('DQOK')
                except ValueError:  # not in list
                    pass
        else:
            idq_msg = ("iDQ false alarm probabilities at both H1 and L1 "
                       "are good (above {} threshold). "
                       "Minimum FAP is {}. ").format(
                           app.conf['idq_fap_thresh'],
                           json.dumps(idq_faps_readable)[1:-1])
    else:
        idq_msg = "iDQ false alarm probabilities unknown. "
    gracedb.upload.delay(
        None, None, graceid, idq_msg + prepost_msg, ['data_quality'])

    # Labeling INJ to GraceDB
    if False in active_inj_states.values():
        # Label 'INJ' if injection found in active IFOs
        gracedb.create_label('INJ', graceid)
        # Add labels to return value to avoid querying GraceDB again.
        event = dict(event, labels=event.get('labels', []) + ['INJ'])
    if False in inj_states.values():
        # Write all found injections into GraceDB log
        injs = [k for k, v in inj_states.items() if v is False]
        inj_fmt = "Injection found.\n{}\n"
        inj_msg = inj_fmt.format(
            generate_table('Injection bits', [], injs, []))
    elif all(inj_states.values()) and len(inj_states.values()) > 0:
        inj_msg = 'No HW injections found. '
        gracedb.remove_label('INJ', graceid)
        event = dict(event, labels=list(event.get('labels', [])))
        try:
            event['labels'].remove('INJ')
        except ValueError:  # not in list
            pass
    else:
        inj_msg = 'Injection state unknown. '
    gracedb.upload.delay(
        None, None, graceid, inj_msg + prepost_msg, ['data_quality'])

    # Determining overall_dq_active_state
    if None in active_dq_states.values() or len(
            active_dq_states.values()) == 0:
        overall_dq_active_state = None
    elif False in active_dq_states.values():
        overall_dq_active_state = False
    elif all(active_dq_states.values()):
        overall_dq_active_state = True
    goods = [k for k, v in dq_states.items() if v is True]
    bads = [k for k, v in dq_states.items() if v is False]
    unknowns = [k for k, v in dq_states.items() if v is None]
    fmt = "Detector state for active instruments is {}.\n{}"
    msg = fmt.format(
        {None: 'unknown', False: 'bad', True: 'good'}[overall_dq_active_state],
        generate_table('Data quality bits', goods, bads, unknowns)
    )
    if app.conf['uses_gatedhoft'][pipeline]:
        gate_msg = ('Pipeline {} uses gated h(t),'
                    ' LIGO DMT-DQ_VECTOR not checked.').format(pipeline)
    else:
        gate_msg = ''

    # Labeling DQOK/DQV to GraceDB
    gracedb.upload.delay(
        None, None, graceid, msg + prepost_msg + gate_msg, ['data_quality'])
    if overall_dq_active_state is True:
        state = "pass"
        gracedb.remove_label('DQV', graceid)
        gracedb.create_label('DQOK', graceid)
        # Add labels to return value to avoid querying GraceDB again.
        event = dict(event, labels=event.get('labels', []) + ['DQOK'])
        try:
            event['labels'].remove('DQV')
        except ValueError:  # not in list
            pass
    elif overall_dq_active_state is False:
        state = "fail"
        gracedb.remove_label('DQOK', graceid)
        gracedb.create_label('DQV', graceid)
        # Add labels to return value to avoid querying GraceDB again.
        event = dict(event, labels=event.get('labels', []) + ['DQV'])
        try:
            event['labels'].remove('DQOK')
        except ValueError:  # not in list
            pass
    else:
        state = "unknown"

    # Create and upload DQR-compatible json
    state_summary = '{} {} {}'.format(inj_msg, idq_msg, msg)
    if state == "unknown":
        json_state = "error"
    else:
        json_state = state
    file = dqr_json(json_state, state_summary)
    filename = 'gwcelerydetcharcheckvectors-{}.json'.format(graceid)
    message = "DQR-compatible json generated from check_vectors results"
    gracedb.upload.delay(
        json.dumps(file), filename, graceid, message)

    return event



"""Tasks that comprise the alert orchestrator.

The orchestrator is responsible for the vetting and annotation workflow to
produce preliminary, initial, and update alerts for gravitational-wave event
candidates.
"""
import json
import re

from celery import chain, group

from ..import app
from . import alerts
from . import bayestar
from . import circulars
from .core import identity, get_first
from . import detchar
from . import em_bright
from . import external_skymaps
from . import gcn
from . import gracedb
from . import inference
from . import igwn_alert
from . import p_astro
from . import rrt_utils
from . import skymaps
from . import superevents

from ligo.rrt_chat import channel_creation


@igwn_alert.handler('superevent',
                    'mdc_superevent',
                    shared=False)
def handle_superevent(alert):
    """Schedule annotations for new superevents.

    After waiting for a time specified by the
    :obj:`~gwcelery.conf.orchestrator_timeout` configuration variable for the
    choice of preferred event to settle down, this task performs data quality
    checks with :meth:`gwcelery.tasks.detchar.check_vectors` and calls
    :meth:`~gwcelery.tasks.orchestrator.earlywarning_preliminary_alert` to send
    a preliminary notice.
    """
    superevent_id = alert['uid']
    # launch PE and detchar based on new type superevents
    if alert['alert_type'] == 'new':
        (
            _get_preferred_event.si(superevent_id).set(
                countdown=app.conf['pe_timeout']
            )
            |
            group(
                _get_lowest_far.si(superevent_id),
                gracedb.get_event.s()
            )
            |
            parameter_estimation.s(superevent_id)
        ).apply_async()

        # run check_vectors. Create and upload omegascans
        group(
            detchar.omegascan.si(alert['object']['t_0'], superevent_id),

            detchar.check_vectors.si(
                alert['object']['preferred_event_data'],
                superevent_id,
                alert['object']['t_start'],
                alert['object']['t_end']
            )
        ).delay()

    elif alert['alert_type'] == 'label_added':
        label_name = alert['data']['name']
        query = f'superevent: {superevent_id} group: CBC Burst'
        if alert['object']['category'] == 'MDC':
            query += ' MDC'
        elif alert['object']['category'] == 'Test':
            query += ' Test'

        # launch less-significant preliminary alerts on LOW_SIGNIF_LOCKED
        if label_name == superevents.FROZEN_LABEL:
            # don't launch if EARLY_WARNING or SIGNIF_LOCKED is present
            skipping_labels = {
                superevents.SIGNIFICANT_LABEL,
                superevents.EARLY_WARNING_LABEL
            }.intersection(alert['object']['labels'])
            if skipping_labels:
                gracedb.upload.delay(
                    None, None, superevent_id,
                    "The superevent already has a significant/EW event, "
                    "skipping launching less-significant alert"
                )
                return
            (
                gracedb.upload.s(
                    None,
                    None,
                    superevent_id,
                    "Automated DQ check before sending less-significant "
                    "preliminary alert. New results supersede old results.",
                    tags=['data_quality']
                )
                |
                detchar.check_vectors.si(
                    alert['object']['preferred_event_data'],
                    superevent_id,
                    alert['object']['t_start'],
                    alert['object']['t_end']
                )
                |
                earlywarning_preliminary_alert.s(
                    alert, alert_type='less-significant')
            ).apply_async()

        # launch significant alert on SIGNIF_LOCKED
        elif label_name == superevents.SIGNIFICANT_LABEL:
            # ensure superevent is locked before starting alert workflow
            if superevents.FROZEN_LABEL not in alert['object']['labels']:
                gracedb.create_label(superevents.FROZEN_LABEL, superevent_id)

            (
                gracedb.get_events.si(query)
                |
                superevents.select_preferred_event.s()
                |
                _update_superevent_and_return_event_dict.s(superevent_id)
                |
                _leave_log_message_and_return_event_dict.s(
                    superevent_id,
                    "Superevent cleaned up after significant event. "
                )
                |
                _leave_log_message_and_return_event_dict.s(
                    superevent_id,
                    "Automated DQ check before sending significant alert. "
                    "New results supersede old results.",
                    tags=['data_quality']
                )
                |
                detchar.check_vectors.s(
                    superevent_id,
                    alert['object']['t_start'],
                    alert['object']['t_end']
                )
                |
                earlywarning_preliminary_alert.s(
                    alert, alert_type='preliminary')
            ).apply_async()

        # launch second preliminary on GCN_PRELIM_SENT
        elif label_name == 'GCN_PRELIM_SENT':
            (
                identity.si().set(
                    # https://git.ligo.org/emfollow/gwcelery/-/issues/478
                    # FIXME: remove this task once https://github.com/celery/celery/issues/7851 is resolved  # noqa: E501
                    countdown=app.conf['superevent_clean_up_timeout']
                )
                |
                gracedb.get_events.si(query)
                |
                superevents.select_preferred_event.s()
                |
                _update_superevent_and_return_event_dict.s(superevent_id)
                |
                group(
                    _leave_log_message_and_return_event_dict.s(
                        superevent_id,
                        "Superevent cleaned up after first preliminary alert"
                    ),

                    gracedb.create_label.si('DQR_REQUEST', superevent_id)
                )
                |
                get_first.s()
                |
                earlywarning_preliminary_alert.s(
                    alert, alert_type='preliminary')
            ).apply_async()

            # set pipeline preferred events
            # FIXME: Ideally this should combined with the previous canvas.
            # However, incorporating that group prevents canvas from executing
            # maybe related to https://github.com/celery/celery/issues/7851
            (
                gracedb.get_events.si(query)
                |
                superevents.select_pipeline_preferred_event.s()
                |
                _set_pipeline_preferred_events.s(superevent_id)
            ).apply_async(countdown=app.conf['superevent_clean_up_timeout'])

        elif label_name == 'LOW_SIGNIF_PRELIM_SENT':
            # similar workflow as the GCN_PRELIM_SENT
            # except block by condition evaluated at the end of timeout
            _revise_and_send_second_less_significant_alert.si(
                    alert, query, superevent_id,
            ).apply_async(countdown=app.conf['superevent_clean_up_timeout'])

        elif label_name == superevents.EARLY_WARNING_LABEL:
            if superevents.SIGNIFICANT_LABEL in alert['object']['labels']:
                # stop if full BW significant event already present
                gracedb.upload.delay(
                    None, None, superevent_id,
                    "Superevent superseded by full BW event, skipping EW."
                )
                return
            # start the EW alert pipeline; is blocked by SIGNIF_LOCKED
            # ensure superevent is locked before starting pipeline
            if superevents.FROZEN_LABEL not in alert['object']['labels']:
                gracedb.create_label(superevents.FROZEN_LABEL, superevent_id)

            (
                gracedb.get_events.si(query)
                |
                superevents.select_preferred_event.s()
                |
                _update_superevent_and_return_event_dict.s(superevent_id)
                |
                _leave_log_message_and_return_event_dict.s(
                    superevent_id,
                    "Superevent cleaned up before sending EW alert."
                )
                |
                earlywarning_preliminary_alert.s(
                    alert, alert_type='earlywarning')
            ).apply_async()

        # launch initial/retraction alert on ADVOK/ADVNO
        elif label_name == 'ADVOK':
            initial_alert((None, None, None), alert)
        elif label_name == 'ADVNO':
            retraction_alert(alert)
        elif label_name == 'ADVREQ':
            if app.conf['create_mattermost_channel']:
                _create_mattermost_channel.si(superevent_id).delay()

    # check DQV label on superevent, run check_vectors if required
    elif alert['alert_type'] == 'event_added':
        # FIXME Check if this should be changed to 'update' alert_types instead
        # of 'event_added'. 'event_added' seems like it's just a new event in
        # the superevent window, not necessarily an event that should be
        # promoted to preferred event
        start = alert['data']['t_start']
        end = alert['data']['t_end']

        if 'DQV' in gracedb.get_labels(superevent_id):
            (
                detchar.check_vectors.s(
                    alert['object']['preferred_event_data'],
                    superevent_id,
                    start,
                    end
                )
                |
                _update_if_dqok.si(
                    superevent_id,
                    alert['object']['preferred_event_data']['graceid']
                )
            ).apply_async()


@igwn_alert.handler('cbc_gstlal',
                    'cbc_spiir',
                    'cbc_pycbc',
                    'cbc_mbta',
                    shared=False)
def handle_cbc_event(alert):
    """Perform annotations for CBC events that depend on pipeline-specific
    matched-filter parameter estimates.

    Notes
    -----
    This IGWN alert message handler is triggered by a new upload or by updates
    that include the file ``pipeline.p_astro.json``. If also generates
    pipeline.p_astro.json information for pipelines that do not provide
    such information.

    The table below lists which files are created as a result of a new upload,
    and which tasks generate them.

    ============================== ==================================================
    File                           Task
    ============================== ==================================================
    ``bayestar.multiorder.fits``   :meth:`gwcelery.tasks.bayestar.localize`
    ``em_bright.json``             :meth:`gwcelery.tasks.em_bright.source_properties`
    ``pipeline.p_astro.json``      :meth:`gwcelery.tasks.p_astro.compute_p_astro`
    ============================== ==================================================

    """  # noqa: E501
    graceid = alert['uid']
    pipeline = alert['object']['pipeline'].lower()
    search = alert['object']['search'].lower()
    priority = 0 if superevents.should_publish(alert['object']) else 1

    # Pipelines that use the GWCelery p-astro method
    # - spiir (all searches)
    # - pycbc for EarlyWarning search
    # - periodic MDC generated by first-two-years (based on gstlal)
    # FIXME: Remove this once all pipelines compute their own p-astro
    pipelines_stock_p_astro = {('spiir', 'earlywarning'),
                               ('pycbc', 'earlywarning'),
                               ('gstlal', 'mdc')}

    # em_bright and p_astro calculation
    if alert['alert_type'] == 'new':
        instruments = superevents.get_instruments_in_ranking_statistic(
            alert['object'])
        extra_attributes = alert['object']['extra_attributes']
        snr = superevents.get_snr(alert['object'])
        far = alert['object']['far']
        mass1 = extra_attributes['SingleInspiral'][0]['mass1']
        mass2 = extra_attributes['SingleInspiral'][0]['mass2']
        chi1 = extra_attributes['SingleInspiral'][0]['spin1z']
        chi2 = extra_attributes['SingleInspiral'][0]['spin2z']

        (
            em_bright.source_properties.si(mass1, mass2, chi1, chi2, snr)
            |
            gracedb.upload.s(
                'em_bright.json', graceid,
                'em bright complete', ['em_bright', 'public']
            )
            |
            gracedb.create_label.si('EMBRIGHT_READY', graceid)
        ).apply_async(priority=priority)

        # p_astro calculation for pipelines that does not provide a
        # stock p_astro (upload pipeline.p_astro.json)
        if (pipeline, search) in pipelines_stock_p_astro:
            (
                p_astro.compute_p_astro.s(snr,
                                          far,
                                          mass1,
                                          mass2,
                                          pipeline,
                                          instruments)
                |
                gracedb.upload.s(
                    f'{pipeline}.p_astro.json', graceid,
                    'p_astro computation complete', ['p_astro', 'public']
                )
                |
                gracedb.create_label.si('PASTRO_READY', graceid)
            ).apply_async(priority=priority)

        # Start BAYESTAR for all CBC pipelines.
        (
            gracedb.download.s('coinc.xml', graceid)
            |
            bayestar.localize.s(graceid)
            |
            gracedb.upload.s(
                'bayestar.multiorder.fits', graceid,
                'sky localization complete', ['sky_loc', 'public']
            )
            |
            gracedb.create_label.si('SKYMAP_READY', graceid)
        ).apply_async(priority=priority)


@igwn_alert.handler('burst_olib',
                    'burst_cwb',
                    'burst_mly',
                    shared=False)
def handle_burst_event(alert):
    """Perform annotations for burst events that depend on pipeline-specific
    """  # noqa: E501
    graceid = alert['uid']
    priority = 0 if superevents.should_publish(alert['object']) else 1

    if alert['alert_type'] != 'log':
        return

    filename = alert['data']['filename']

    # Pipeline is uploading a flat resultion skymap file
    # Converting to a multiirder ones with proper name.
    # FIXME: Remove block when CWB starts to upload skymaps
    # in multiorder format
    if filename.endswith('.fits.gz'):
        new_filename = filename.replace('.fits.gz', '.multiorder.fits')
        flatten_msg = (
            'Multi-resolution fits file created from '
            '<a href="/api/events/{graceid}/files/'
            '{filename}">{filename}</a>').format(
            graceid=graceid, filename=filename)
        tags = ['sky_loc', 'lvem', 'public']
        (
            gracedb.download.si(filename, graceid)
            |
            skymaps.unflatten.s(new_filename)
            |
            gracedb.upload.s(
                    new_filename, graceid, flatten_msg, tags)
            |
            gracedb.create_label.si('SKYMAP_READY', graceid)
        ).apply_async(priority=priority)


@igwn_alert.handler('superevent',
                    'mdc_superevent',
                    shared=False)
def handle_posterior_samples(alert):
    """Generate multi-resolution and flat-resolution fits files and skymaps
    from an uploaded HDF5 file containing posterior samples.
    """
    if alert['alert_type'] != 'log' or \
            not alert['data']['filename'].endswith('.posterior_samples.hdf5'):
        return
    superevent_id = alert['uid']
    filename = alert['data']['filename']
    info = '{} {}'.format(alert['data']['comment'], filename)
    prefix, _ = filename.rsplit('.posterior_samples.')
    skymap_filename = f'{prefix}.multiorder.fits'
    labels = ['pe', 'sky_loc', 'public']

    (
        gracedb.download.si(filename, superevent_id)
        |
        skymaps.skymap_from_samples.s()
        |
        group(
            skymaps.annotate_fits.s(
                skymap_filename, superevent_id, labels
            ),

            gracedb.upload.s(
                skymap_filename, superevent_id,
                'Multiresolution fits file generated from "{}"'.format(info),
                labels
            )
        )
    ).delay()

    # em_bright from LALInference posterior samples
    (
        gracedb.download.si(filename, superevent_id)
        |
        em_bright.em_bright_posterior_samples.s()
        |
        gracedb.upload.s(
            '{}.em_bright.json'.format(prefix), superevent_id,
            'em-bright computed from "{}"'.format(info)
        )
    ).delay()


@app.task(bind=True, shared=False)
def _create_mattermost_channel(self, superevent_id):
    """
    Creates a mattermost channel when ADVREQ label is applied and
    posts a cooresponding gracedb link of that event in the channel

    Channel name : O4 RRT {superevent_id}

    Parameters:
    ------------
    superevent_id: str
        The superevent id
    """
    gracedb_url = self.app.conf['gracedb_host']
    channel_creation.rrt_channel_creation(
        superevent_id, gracedb_url)


@app.task(shared=False)
def _set_pipeline_preferred_events(pipeline_event, superevent_id):
    """Return group for setting pipeline preferred event using
    :meth:`gracedb.add_pipeline_preferred_event`.

    Parameters
    ----------
    pipeline_event: dict
        {pipeline: event_dict} key value pairs, returned by
        :meth:`superevents.select_pipeline_preferred_event`.

    superevent_id: str
        The superevent id
    """
    return group(
        gracedb.add_pipeline_preferred_event(superevent_id,
                                             event['graceid'])
        for event in pipeline_event.values()
    )


@app.task(shared=False, ignore_result=True)
def _update_if_dqok(superevent_id, event_id):
    """Update `preferred_event` of `superevent_id` to `event_id` if `DQOK`
    label has been applied.
    """
    if 'DQOK' in gracedb.get_labels(superevent_id):
        gracedb.update_superevent(superevent_id, preferred_event=event_id)
        gracedb.create_log.delay(
            f'DQOK applied based on new event {event_id}', superevent_id)


@gracedb.task(shared=False)
def _get_preferred_event(superevent_id):
    """Determine preferred event for a superevent by querying GraceDB.

    This works just like :func:`gwcelery.tasks.gracedb.get_superevent`, except
    that it returns only the preferred event, and not the entire GraceDB JSON
    response.
    """
    # FIXME: remove ._orig_run when this bug is fixed:
    # https://github.com/getsentry/sentry-python/issues/370
    return gracedb.get_superevent._orig_run(superevent_id)['preferred_event']


@gracedb.task(shared=False)
def _create_voevent(classification, *args, **kwargs):
    r"""Create a VOEvent record from an EM bright JSON file.

    Parameters
    ----------
    classification : tuple, None
        A collection of JSON strings, generated by
        :meth:`gwcelery.tasks.em_bright.source_properties` and
        :meth:`gwcelery.tasks.p_astro.compute_p_astro` or
        content of ``{gstlal,mbta}.p_astro.json`` uploaded
        by {gstlal,mbta} respectively; or None
    \*args
        Additional positional arguments passed to
        :meth:`gwcelery.tasks.gracedb.create_voevent`.
    \*\*kwargs
        Additional keyword arguments passed to
        :meth:`gwcelery.tasks.gracedb.create_voevent`.

    Returns
    -------
    str
        The filename of the newly created VOEvent.

    """
    kwargs = dict(kwargs)

    if classification is not None:
        # Merge source classification and source properties into kwargs.
        for text in classification:
            # Ignore filenames, only load dict in bytes form
            if text is not None:
                kwargs.update(json.loads(text))

    # FIXME: These keys have differ between em_bright.json
    # and the GraceDB REST API.
    try:
        kwargs['ProbHasNS'] = kwargs.pop('HasNS')
    except KeyError:
        pass

    try:
        kwargs['ProbHasRemnant'] = kwargs.pop('HasRemnant')
    except KeyError:
        pass

    skymap_filename = kwargs.get('skymap_filename')
    if skymap_filename is not None:
        skymap_type = re.sub(
            r'(\.multiorder)?\.fits(\..+)?(,[0-9]+)?$', '', skymap_filename)
        kwargs.setdefault('skymap_type', skymap_type)

    # FIXME: remove ._orig_run when this bug is fixed:
    # https://github.com/getsentry/sentry-python/issues/370
    return gracedb.create_voevent._orig_run(*args, **kwargs)


@app.task(shared=False)
def _create_label_and_return_filename(filename, label, graceid):
    gracedb.create_label.delay(label, graceid)
    return filename


@app.task(shared=False)
def _leave_log_message_and_return_event_dict(event, superevent_id,
                                             message, **kwargs):
    """Wrapper around :meth:`gracedb.update_superevent`
    that returns the event dictionary.
    """
    gracedb.upload.delay(None, None, superevent_id, message, **kwargs)
    return event


@gracedb.task(shared=False)
def _update_superevent_and_return_event_dict(event, superevent_id):
    """Wrapper around :meth:`gracedb.update_superevent`
    that returns the event dictionary.
    """
    gracedb.update_superevent(superevent_id,
                              preferred_event=event['graceid'])
    return event


@gracedb.task(shared=False)
def _proceed_if_not_blocked_by(files, superevent_id, block_by):
    """Return files in case the superevent does not have labels `block_by`

    Parameters
    ----------
    files : tuple
        List of files
    superevent_id : str
        The superevent id corresponding to files
    block_by : set
        Set of blocking labels. E.g. `{'ADVOK', 'ADVNO'}`
    """
    superevent_labels = gracedb.get_labels(superevent_id)
    blocking_labels = block_by.intersection(superevent_labels)
    if blocking_labels:
        gracedb.upload.delay(
            None, None, superevent_id,
            f"Blocking automated notice due to labels {blocking_labels}"
        )
        return None
    else:
        return files


@gracedb.task(shared=False)
def _revise_and_send_second_less_significant_alert(alert, query,
                                                   superevent_id):
    superevent_labels = gracedb.get_labels(superevent_id)
    blocking_labels = {
        'ADVREQ', 'ADVOK', 'ADVNO',
        superevents.SIGNIFICANT_LABEL,
        superevents.EARLY_WARNING_LABEL,
    }
    if blocking_labels.intersection(superevent_labels):
        return

    (
        gracedb.get_events.si(query)
        |
        superevents.select_preferred_event.s()
        |
        _update_superevent_and_return_event_dict.s(superevent_id)
        |
        _leave_log_message_and_return_event_dict.s(
            superevent_id,
            "Superevent cleaned up before second less-significant alert"
        )
        |
        earlywarning_preliminary_alert.s(
            alert, alert_type='less-significant')
    ).delay()


@app.task(shared=False)
def _annotate_fits_and_return_input(input_list, superevent_id):
    """Unpack the output of the skymap, embright, p-astro download group in the
    beginning of the
    :meth:`~gwcelery.tasks.orchestartor.earlywarning_preliminary_alert` canvas
    and call :meth:`~gwcelery.tasks.skymaps.annotate_fits`.


    Parameters
    ----------
    input_list : list
        The output of the group that downloads the skymap, embright, and
        p-astro files. This list is in the form [skymap, skymap_filename],
        [em_bright, em_bright_filename], [p_astro, p_astro_filename], though
        the em-bright and p-astro lists can be populated by Nones
    superevent_id : str
        A list of the sky map, em_bright, and p_astro filenames.
    """

    [skymap, skymap_filename], *_ = input_list

    skymaps.annotate_fits.delay(
        skymap,
        skymap_filename,
        superevent_id,
        ['sky_loc', 'public']
    )

    return input_list


@app.task(shared=False)
def _unpack_args_and_send_earlywarning_preliminary_alert(input_list, alert,
                                                         alert_type):
    """Unpack the output of the skymap, embright, p-astro download group in the
    beginning of the
    :meth:`~gwcelery.tasks.orchestartor.earlywarning_preliminary_alert` canvas
    and call
    :meth:`gwcelery.tasks.orchestrator.earlywarning_preliminary_initial_update_alert`.


    Parameters
    ----------
    input_list : list
        The output of the group that downloads the skymap, embright, and
        p-astro files. This list is in the form [skymap, skymap_filename],
        [em_bright, em_bright_filename], [p_astro, p_astro_filename], though
        the em-bright and p-astro lists can be populated by Nones
    alert : dict
        IGWN-Alert dictionary
    alert_type : str
        alert_type passed to
        :meth:`earlywarning_preliminary_initial_update_alert`
    """
    if input_list is None:  # alert is blocked by blocking labels
        return

    [skymap, skymap_filename], [em_bright, em_bright_filename], \
        [p_astro, p_astro_filename] = input_list

    # Update to latest state after downloading files
    superevent = gracedb.get_superevent(alert['object']['superevent_id'])

    earlywarning_preliminary_initial_update_alert.delay(
        [skymap_filename, em_bright_filename, p_astro_filename],
        superevent, alert_type,
        filecontents=[skymap, em_bright, p_astro]
    )


@app.task(ignore_result=True, shared=False)
def earlywarning_preliminary_alert(event, alert, alert_type='preliminary',
                                   initiate_voevent=True):
    """Produce a preliminary alert by copying any sky maps.

    This consists of the following steps:

    1.   Copy any sky maps and source classification from the preferred event
         to the superevent.
    2.   Create standard annotations for sky maps including all-sky plots by
         calling :meth:`gwcelery.tasks.skymaps.annotate_fits`.
    3.   Create a preliminary VOEvent.
    4.   Send the VOEvent to GCN and notices to SCiMMA and GCN.
    5.   Apply the GCN_PRELIM_SENT or LOW_SIGNIF_PRELIM_SENT
         depending on the significant or less-significant alert
         respectively.
    6.   Create and upload a GCN Circular draft.
    """
    priority = 0 if superevents.should_publish(event) else 1
    preferred_event_id = event['graceid']
    superevent_id = alert['uid']

    if event['group'] == 'CBC':
        skymap_filename = 'bayestar.multiorder.fits'
        p_astro_filename = event['pipeline'].lower() + '.p_astro.json'
    elif event['group'] == 'Burst':
        skymap_filename = event['pipeline'].lower() + '.multiorder.fits'
        p_astro_filename = None
    else:
        raise NotImplementedError(
            'Valid skymap required for preliminary alert'
        )

    # Determine if the event should be made public.
    is_publishable = (superevents.should_publish(
                      event, significant=alert_type != 'less-significant') and
                      {'DQV', 'INJ'}.isdisjoint(
                      event['labels']))

    # Download files from events and upload to superevent with relevant
    # annotations. Pass file contents down the chain so alerts task doesn't
    # need to download them again.
    # Note: this is explicitly made a chain to fix an issue described in #464.
    canvas = chain(
        group(
            gracedb.download.si(skymap_filename, preferred_event_id)
            |
            group(
                identity.s(),

                gracedb.upload.s(
                    skymap_filename,
                    superevent_id,
                    message='Localization copied from {}'.format(
                        preferred_event_id),
                    tags=['sky_loc', 'public']
                )
                |
                _create_label_and_return_filename.s('SKYMAP_READY',
                                                    superevent_id)
            ),

            (
                gracedb.download.si('em_bright.json', preferred_event_id)
                |
                group(
                    identity.s(),

                    gracedb.upload.s(
                        'em_bright.json',
                        superevent_id,
                        message='Source properties copied from {}'.format(
                            preferred_event_id),
                        tags=['em_bright', 'public']
                    )
                    |
                    _create_label_and_return_filename.s('EMBRIGHT_READY',
                                                        superevent_id)
                )
            ) if event['group'] == 'CBC' else
            identity.s([None, None]),

            (
                gracedb.download.si(p_astro_filename, preferred_event_id)
                |
                group(
                    identity.s(),

                    gracedb.upload.s(
                        p_astro_filename,
                        superevent_id,
                        message='Source classification copied from {}'.format(
                            preferred_event_id),
                        tags=['p_astro', 'public']
                    )
                    |
                    _create_label_and_return_filename.s('PASTRO_READY',
                                                        superevent_id)
                )
            ) if event['group'] == 'CBC' else
            identity.s([None, None])
        )
        |
        # Need annotate skymap task in body of chord instead of header because
        # this task simply calls another task, which is to be avoided in chord
        # headers. Note that any group that chains to a task is automatically
        # upgraded to a chord.
        _annotate_fits_and_return_input.s(superevent_id)
    )

    # Switch for disabling all but MDC alerts.
    if app.conf['only_alert_for_mdc']:
        if event.get('search') != 'MDC':
            canvas |= gracedb.upload.si(
                None, None, superevent_id,
                ("Skipping alert because gwcelery has been configured to only"
                 " send alerts for MDC events."))
            canvas.apply_async(priority=priority)
            return

    # Send notice and upload GCN circular draft for online events.
    if is_publishable and initiate_voevent:
        # presence of advocate action blocks significant prelim alert
        # presence of adv action or significant event blocks EW alert
        # presence of adv action or significant event or EW event blocks
        # less significant alert
        blocking_labels = (
            {'ADVOK', 'ADVNO'} if alert_type == 'preliminary'
            else
            {superevents.SIGNIFICANT_LABEL, 'ADVOK', 'ADVNO'}
            if alert_type == 'earlywarning'
            else
            {superevents.EARLY_WARNING_LABEL, superevents.SIGNIFICANT_LABEL,
             'ADVOK', 'ADVNO'}
            if alert_type == 'less-significant'
            else
            set()
        )
        canvas |= (
            _proceed_if_not_blocked_by.s(superevent_id, blocking_labels)
            |
            _unpack_args_and_send_earlywarning_preliminary_alert.s(
                alert, alert_type
            )
        )

    canvas.apply_async(priority=priority)


@gracedb.task(shared=False)
def _get_lowest_far(superevent_id):
    """Obtain the lowest FAR of the events in the target superevent."""
    # FIXME: remove ._orig_run when this bug is fixed:
    # https://github.com/getsentry/sentry-python/issues/370
    return min(gracedb.get_event._orig_run(gid)['far'] for gid in
               gracedb.get_superevent._orig_run(superevent_id)["gw_events"])


@app.task(ignore_result=True, shared=False)
def parameter_estimation(far_event, superevent_id):
    """Parameter Estimation with Bilby and RapidPE-RIFT. Parameter estimation
    runs are triggered for CBC triggers which pass the FAR threshold and are
    not mock uploads. For those which do not pass these criteria, this task
    uploads messages explaining why parameter estimation is not started.
    """
    far, event = far_event
    threshold = (app.conf['significant_alert_far_threshold']['cbc'] /
                 app.conf['significant_alert_trials_factor']['cbc'])
    if far > threshold:
        gracedb.upload.delay(
            filecontents=None, filename=None,
            graceid=superevent_id,
            message='Parameter estimation will not start since FAR is larger '
                    'than the PE threshold, {}  Hz.'.format(threshold),
            tags='pe'
        )
    elif event['group'] != 'CBC':
        gracedb.upload.delay(
            filecontents=None, filename=None,
            graceid=superevent_id,
            message='Parameter estimation will not start since this is not '
                    'CBC but {}.'.format(event['group']),
            tags='pe'
        )
    elif event['search'] == 'MDC':
        gracedb.upload.delay(
            filecontents=None, filename=None,
            graceid=superevent_id,
            message='Parameter estimation will not start since parameter '
                    'estimation is disabled for mock uploads.',
            tags='pe'
        )
    else:
        canvas = inference.query_data.s(event['gpstime']).on_error(
            inference.upload_no_frame_files.s(superevent_id)
        )
        pe_pipelines = ['bilby']
        # Currently rapidpe works only for gstlal triggers
        if event['pipeline'] == 'gstlal':
            pe_pipelines += ['rapidpe']
        canvas |= group(
            inference.start_pe.s(event, superevent_id, p)
            for p in pe_pipelines)
        canvas.apply_async()


@gracedb.task(shared=False)
def earlywarning_preliminary_initial_update_alert(
    filenames,
    superevent,
    alert_type,
    filecontents=None
):
    """
    Create a canvas that sends an earlywarning, preliminary, initial, or update
    notice.

    Parameters
    ----------
    filenames : tuple
        A list of the sky map, em_bright, and p_astro filenames.
    superevent : dict
        The superevent dictionary, typically obtained from an IGWN Alert or
        from querying GraceDB.
    alert_type : {'less-significant', 'earlywarning', 'preliminary', 'initial', 'update'}  # noqa: E501
        The alert type.

    Notes
    -----
    Tasks that call this function should be decorated with
    :obj:`gwcelery.tasks.gracedb.task` rather than :obj:`gwcelery.app.task` so
    that a synchronous call to :func:`gwcelery.tasks.gracedb.get_log` is
    retried in the event of GraceDB API failures. If `EM_COINC` is in labels
    will create a RAVEN circular.

    """
    labels = superevent['labels']
    superevent_id = superevent['superevent_id']

    if 'INJ' in labels:
        return

    if filecontents:
        assert alert_type in {
            'less-significant', 'earlywarning', 'preliminary'}

    skymap_filename, em_bright_filename, p_astro_filename = filenames
    combined_skymap_filename = None
    combined_skymap_needed = False
    skymap_needed = (skymap_filename is None)
    em_bright_needed = (em_bright_filename is None)
    p_astro_needed = (p_astro_filename is None)
    raven_coinc = ('RAVEN_ALERT' in labels and bool(superevent['em_type']))
    if raven_coinc:
        ext_labels = gracedb.get_labels(superevent['em_type'])
        combined_skymap_needed = \
            {"RAVEN_ALERT", "COMBINEDSKYMAP_READY"}.issubset(set(ext_labels))

    if skymap_needed or em_bright_needed or p_astro_needed or \
            combined_skymap_needed:
        for message in gracedb.get_log(superevent_id):
            t = message['tag_names']
            f = message['filename']
            v = message['file_version']
            fv = '{},{}'.format(f, v)
            if not f:
                continue
            if skymap_needed \
                    and {'sky_loc', 'public'}.issubset(t) \
                    and f.endswith('.multiorder.fits') \
                    and 'combined' not in f:
                skymap_filename = fv
            if em_bright_needed \
                    and 'em_bright' in t \
                    and f.endswith('.json'):
                em_bright_filename = fv
            if p_astro_needed \
                    and 'p_astro' in t \
                    and f.endswith('.json'):
                p_astro_filename = fv
            if combined_skymap_needed \
                    and {'sky_loc', 'ext_coinc'}.issubset(t) \
                    and f.startswith('combined-ext.') \
                    and 'fit' in f:
                combined_skymap_filename = fv

    if combined_skymap_needed:
        # for every alert, copy combined sky map over if applicable
        # FIXME: use file inheritance once available
        ext_id = superevent['em_type']
        if combined_skymap_filename:
            # If previous sky map, increase version by 1
            combined_skymap_filename_base, v = \
                combined_skymap_filename.split(',')
            v = str(int(v) + 1)
            combined_skymap_filename = \
                combined_skymap_filename_base + ',' + v
        else:
            combined_skymap_filename_base = \
                (external_skymaps.COMBINED_SKYMAP_FILENAME_MULTIORDER
                 if '.multiorder.fits' in skymap_filename else
                 external_skymaps.COMBINED_SKYMAP_FILENAME_FLAT)
            combined_skymap_filename = combined_skymap_filename_base + ',0'
        message = 'Combined LVK-external sky map copied from {0}'.format(
            ext_id)
        message_png = (
            'Mollweide projection of <a href="/api/events/{se_id}/files/'
            '{filename}">{filename}</a>, copied from {ext_id}').format(
               se_id=superevent_id,
               ext_id=ext_id,
               filename=combined_skymap_filename)

        combined_skymap_canvas = group(
            gracedb.download.si(combined_skymap_filename_base, ext_id)
            |
            gracedb.upload.s(
                combined_skymap_filename_base, superevent_id,
                message, ['sky_loc', 'ext_coinc', 'public'])
            |
            gracedb.create_label.si('COMBINEDSKYMAP_READY', superevent_id),

            gracedb.download.si(external_skymaps.COMBINED_SKYMAP_FILENAME_PNG,
                                ext_id)
            |
            gracedb.upload.s(
                external_skymaps.COMBINED_SKYMAP_FILENAME_PNG, superevent_id,
                message_png, ['sky_loc', 'ext_coinc', 'public']
            )
            |
            # Pass None to download_anor_expose group
            identity.si()
        )

    # circular template not needed for less-significant alerts
    if alert_type in {'earlywarning', 'preliminary', 'initial'}:
        if raven_coinc:
            circular_task = circulars.create_emcoinc_circular.si(superevent_id)
            circular_filename = '{}-emcoinc-circular.txt'.format(alert_type)
            tags = ['em_follow', 'ext_coinc']

        else:
            circular_task = circulars.create_initial_circular.si(superevent_id)
            circular_filename = '{}-circular.txt'.format(alert_type)
            tags = ['em_follow']

        circular_canvas = (
            circular_task
            |
            gracedb.upload.s(
                circular_filename,
                superevent_id,
                'Template for {} GCN Circular'.format(alert_type),
                tags=tags)
        )

    else:
        circular_canvas = identity.si()

    # less-significant alerts have "preliminary" voevent notice type
    alert_type_voevent = 'preliminary' if alert_type == 'less-significant' \
        else alert_type
    # set the significant field in the VOEvent based on
    # less-significant/significant alert.
    # For kafka alerts the analogous field is set in alerts.py.
    # (see comment before defining kafka_alert_canvas)
    voevent_significance = 0 if alert_type == 'less-significant' else 1

    if filecontents and not combined_skymap_filename:
        skymap, em_bright, p_astro = filecontents

        # check high profile and apply label if true
        high_profile_canvas = rrt_utils.check_high_profile.si(
            skymap, em_bright, p_astro, superevent
        )

        download_andor_expose_group = []

        voevent_canvas = _create_voevent.si(
            (em_bright, p_astro),
            superevent_id,
            alert_type_voevent,
            Significant=voevent_significance,
            skymap_filename=skymap_filename,
            internal=False,
            open_alert=True,
            raven_coinc=raven_coinc,
            combined_skymap_filename=combined_skymap_filename
        )
        # kafka alerts have a field called "significant" based on
        # https://dcc.ligo.org/LIGO-G2300151/public
        # The alert_type value passed to alerts.send is used to
        # set this field in the alert dictionary
        kafka_alert_canvas = alerts.send.si(
            (skymap, em_bright, p_astro),
            superevent,
            alert_type,
            raven_coinc=raven_coinc
        )
    else:
        # Download em_bright and p_astro files here for voevent
        download_andor_expose_group = [
            gracedb.download.si(em_bright_filename, superevent_id),
            gracedb.download.si(p_astro_filename, superevent_id),
        ]
        high_profile_canvas = identity.si()

        voevent_canvas = _create_voevent.s(
            superevent_id,
            alert_type_voevent,
            Significant=voevent_significance,
            skymap_filename=skymap_filename,
            internal=False,
            open_alert=True,
            raven_coinc=raven_coinc,
            combined_skymap_filename=combined_skymap_filename
        )

        # The skymap has not been downloaded at this point, so we need to
        # download it before we can assemble the kafka alerts and send them
        kafka_alert_canvas = alerts.download_skymap_and_send_alert.s(
            superevent,
            alert_type,
            skymap_filename=skymap_filename,
            raven_coinc=raven_coinc,
            combined_skymap_filename=combined_skymap_filename
        )

    download_andor_expose_group += [
        gracedb.expose.si(superevent_id),
        *(
            gracedb.create_tag.si(filename, 'public', superevent_id)
            for filename in [
                skymap_filename, em_bright_filename, p_astro_filename
            ]
            if filename is not None
        )
    ]

    voevent_canvas |= group(
        gracedb.download.s(superevent_id)
        |
        gcn.send.s(),

        gracedb.create_tag.s('public', superevent_id)
    )

    if combined_skymap_needed:
        download_andor_expose_group += [combined_skymap_canvas]

    sent_label_canvas = identity.si()
    if alert_type == 'less-significant':
        sent_label_canvas = gracedb.create_label.si(
            'LOW_SIGNIF_PRELIM_SENT',
            superevent_id
        )
    elif alert_type == 'preliminary':
        sent_label_canvas = gracedb.create_label.si(
            'GCN_PRELIM_SENT',
            superevent_id
        )

    canvas = (
        group(download_andor_expose_group)
        |
        group(voevent_canvas, kafka_alert_canvas)
        |
        group(
            sent_label_canvas,
            circular_canvas,
            high_profile_canvas
        )
    )

    canvas.apply_async()


@gracedb.task(ignore_result=True, shared=False)
def initial_alert(filenames, alert):
    """Produce an initial alert.

    This does nothing more than call
    :meth:`~gwcelery.tasks.orchestrator.earlywarning_preliminary_initial_update_alert`
    with ``alert_type='initial'``.

    Parameters
    ----------
    filenames : tuple
        A list of the sky map, em_bright, and p_astro filenames.
    alert : dict
        IGWN-Alert dictionary

    Notes
    -----
    This function is decorated with :obj:`gwcelery.tasks.gracedb.task` rather
    than :obj:`gwcelery.app.task` so that a synchronous call to
    :func:`gwcelery.tasks.gracedb.get_log` is retried in the event of GraceDB
    API failures.

    """
    earlywarning_preliminary_initial_update_alert(
        filenames,
        alert['object'],
        'initial'
    )


@gracedb.task(ignore_result=True, shared=False)
def update_alert(filenames, superevent_id):
    """Produce an update alert.

    This does nothing more than call
    :meth:`~gwcelery.tasks.orchestrator.earlywarning_preliminary_initial_update_alert`
    with ``alert_type='update'``.

    Parameters
    ----------
    filenames : tuple
        A list of the sky map, em_bright, and p_astro filenames.
    superevent_id : str
        The superevent ID.

    Notes
    -----
    This function is decorated with :obj:`gwcelery.tasks.gracedb.task` rather
    than :obj:`gwcelery.app.task` so that a synchronous call to
    :func:`gwcelery.tasks.gracedb.get_log` is retried in the event of GraceDB
    API failures.

    """
    superevent = gracedb.get_superevent._orig_run(superevent_id)
    earlywarning_preliminary_initial_update_alert(
        filenames,
        superevent,
        'update'
    )


@app.task(ignore_result=True, shared=False)
def retraction_alert(alert):
    """Produce a retraction alert."""
    superevent_id = alert['uid']
    group(
        gracedb.expose.si(superevent_id)
        |
        group(
            _create_voevent.si(
                None, superevent_id, 'retraction',
                internal=False,
                open_alert=True
            )
            |
            group(
                gracedb.download.s(superevent_id)
                |
                gcn.send.s(),

                gracedb.create_tag.s('public', superevent_id)
            ),

            alerts.send.si(
                None,
                alert['object'],
                'retraction'
            )
        ),

        circulars.create_retraction_circular.si(superevent_id)
        |
        gracedb.upload.s(
            'retraction-circular.txt',
            superevent_id,
            'Template for retraction GCN Circular',
            tags=['em_follow']
        )
    ).apply_async()



"""Qualitative source properties for CBC events."""
import io
import json
from matplotlib import pyplot as plt

from celery.utils.log import get_task_logger

from ..import app
from . import gracedb, igwn_alert
from .p_astro import _format_prob
from ..util import closing_figures, NamedTemporaryFile


log = get_task_logger(__name__)


@igwn_alert.handler('superevent',
                    'mdc_superevent',
                    shared=False)
def handle(alert):
    """IGWN alert handler to plot and upload a visualization
    of every ``em_bright.json``.
    """
    filename = 'em_bright.json'
    graceid = alert['uid']
    if alert['alert_type'] == 'log' and alert['data']['filename'] == filename:
        (
            gracedb.download.si(filename, graceid)
            |
            plot.s()
            |
            gracedb.upload.s(
                filename.replace('.json', '.png'),
                graceid,
                message=(
                    'Source properties visualization from '
                    '<a href="/api/superevents/{graceid}/files/{filename}">'
                    '{filename}</a>').format(
                        graceid=graceid, filename=filename),
                tags=['em_follow', 'em_bright', 'public']
            )
        ).delay()


@app.task(shared=False)
@closing_figures()
def plot(contents):
    """Make a visualization of the source properties.

    Examples
    --------
    .. plot::
        :include-source:

        >>> from gwcelery.tasks import em_bright
        >>> contents = '{"HasNS": 0.9137, "HasRemnant": 0.0, "HasMassGap": 0.0}'  # noqa E501
        >>> em_bright.plot(contents)
    """
    # Explicitly use a non-interactive Matplotlib backend.
    plt.switch_backend('agg')

    properties = json.loads(contents)
    outfile = io.BytesIO()

    properties = dict(sorted(properties.items(), reverse=True))
    probs, names = list(properties.values()), list(properties.keys())

    with plt.style.context('seaborn-white'):
        fig, ax = plt.subplots(figsize=(3, 1))
        ax.barh(names, probs)
        ax.barh(names, [1.0 - p for p in probs],
                color='lightgray', left=probs)
        for i, prob in enumerate(probs):
            ax.annotate(_format_prob(prob), (0, i), (4, 0),
                        textcoords='offset points', ha='left', va='center')
        ax.set_xlim(0, 1)
        ax.set_xticks([])
        ax.tick_params(left=False)
        for side in ['top', 'bottom', 'right']:
            ax.spines[side].set_visible(False)
        fig.tight_layout()
        fig.savefig(outfile, format='png')
    return outfile.getvalue()


@app.task(shared=False)
def em_bright_posterior_samples(posterior_file_content):
    """Returns the probability of having a NS component and remnant
    using Bilby posterior samples.

    Parameters
    ----------
    posterior_file_content : hdf5 posterior file content

    Returns
    -------
    str
        JSON formatted string storing ``HasNS``, ``HasRemnant``,
        and ``HasMassGap`` probabilities

    Examples
    --------
    >>> em_bright_posterior_samples(GraceDb().files('S190930s',
    ... 'Bilby.posterior_samples.hdf5').read())
    {"HasNS": 0.014904901243599122, "HasRemnant": 0.0, "HasMassGap": 0.0}

    """
    from ligo.em_bright import em_bright
    with NamedTemporaryFile(content=posterior_file_content) as samplefile:
        filename = samplefile.name
        has_ns, has_remnant, has_massgap = em_bright.source_classification_pe(
            filename, num_eos_draws=100, eos_seed=0
        )
    data = json.dumps({
        'HasNS': has_ns,
        'HasRemnant': has_remnant,
        'HasMassGap': has_massgap
    })
    return data


@app.task(shared=False)
def source_properties(mass1, mass2, spin1z, spin2z, snr):
    """Returns the probability of having a NS component, the probability of
    having non-zero disk mass, and the probability of any component being the
    lower mass gap for the detected event.

    Parameters
    ----------
    mass1 : float
        Primary mass in solar masses
    mass2 : float
        Secondary mass in solar masses
    spin1z : float
         Dimensionless primary aligned spin component
    spin2z : float
         Dimensionless secondary aligned spin component
    snr : float
        Signal to noise ratio

    Returns
    -------
    str
        JSON formatted string storing ``HasNS``, ``HasRemnant``,
        and `HasMassGap`` probabilities

    Examples
    --------
    >>> em_bright.source_properties(2.0, 1.0, 0.0, 0.0, 10.)
    '{"HasNS": 1.0, "HasRemnant": 1.0, "HasMassGap"}'
    """
    from ligo.em_bright import em_bright
    p_ns, p_em, p_mg = em_bright.source_classification(
        mass1, mass2, spin1z, spin2z, snr
    )

    data = json.dumps({
        'HasNS': p_ns,
        'HasRemnant': p_em,
        'HasMassGap': p_mg
    })
    return data



import threading

from celery import bootsteps
from celery.concurrency import solo

from .util import get_host_port, get_local_ivo, get_network
from .logging import log
from .signals import voevent_received

__all__ = ('Broadcaster', 'Reactor', 'Receiver')


class VOEventBootStep(bootsteps.ConsumerStep):
    """Generic boot step to limit us to appropriate kinds of workers."""

    def include_if(self, consumer):
        """Only include this bootstep in workers that are configured to listen
        to the ``voevent`` queue.
        """
        return 'voevent' in consumer.app.amqp.queues

    def create(self, consumer):
        if not isinstance(consumer.pool, solo.TaskPool):
            raise RuntimeError(
                'The VOEvent broker only works with the "solo" task pool. '
                'Start the worker with "--queues=voevent --pool=solo".')

    def start(self, consumer):
        log.info('Starting %s', self.name)

    def stop(self, consumer):
        log.info('Stopping %s', self.name)


class Reactor(VOEventBootStep):
    """Run the global Twisted reactor in background thread.

    The Twisted reactor is a global run loop that drives all Twisted services
    and operations. This boot step starts the Twisted reactor in a background
    thread when the Celery consumer starts, and stops the thread when the
    Consumer terminates.
    """

    name = 'Twisted reactor'

    def __init__(self, consumer, **kwargs):
        self._thread = None

    def start(self, consumer):
        super().start(consumer)
        from twisted.internet import reactor
        self._thread = threading.Thread(target=reactor.run, args=(False,),
                                        name='TwistedReactorThread')
        self._thread.start()

    def stop(self, consumer):
        from twisted.internet import reactor

        super().stop(consumer)
        reactor.callFromThread(reactor.stop)
        self._thread.join()


class TwistedService(VOEventBootStep):
    """A generic bootstep to create, start, and stop a Twisted service."""

    requires = VOEventBootStep.requires + (Reactor,)

    def __init__(self, consumer, **kwargs):
        self._service = None

    def create_service(self, consumer):
        raise NotImplementedError

    def start(self, consumer):
        from twisted.internet import reactor

        super().start(consumer)
        self._service = self.create_service(consumer)
        reactor.callFromThread(self._service.startService)

    def stop(self, consumer):
        from twisted.internet import reactor

        super().stop(consumer)
        reactor.callFromThread(self._service.stopService)


class Broadcaster(TwistedService):
    """Comet-based VOEvent broadcaster.

    Run a Comet-based VOEvent broadcaster
    (:class:`comet.protocol.broadcaster.VOEventBroadcasterFactory`). Starts
    after the :class:`~gwcelery.voevent.bootsteps.Reactor` bootstep.

    A few :doc:`configuration options <configuration>` are available:

    * ``voevent_broadcaster_address``: The address to bind to, in
      :samp:`{host}:{port}` format.
    * ``voevent_broadcaster_whitelist``: A list of hostnames, IP addresses, or
       CIDR address ranges from which to accept connections.

    The list of active connections is made available :ref:`inspection
    <celery:worker-inspect>` with the ``gwcelery inspect stats`` command under
    the ``voevent-broker-peers`` key.
    """

    name = 'VOEvent broadcaster'

    def create_service(self, consumer):
        from comet.protocol.broadcaster import VOEventBroadcasterFactory
        from comet.utility import WhitelistingFactory
        from twisted.application.internet import TCPServer

        conf = consumer.app.conf
        local_ivo = get_local_ivo(consumer.app)
        host, port = get_host_port(conf['voevent_broadcaster_address'])
        allow = [get_network(a) for a in conf['voevent_broadcaster_whitelist']]
        conf['voevent_broadcaster_factory'] = self._factory = factory = \
            VOEventBroadcasterFactory(local_ivo, 0)
        if allow:
            factory = WhitelistingFactory(factory, allow, 'subscription')
        return TCPServer(port, factory, interface=host)

    def info(self, consumer):
        try:
            peers = [
                b.transport.getPeer().host for b in self._factory.broadcasters]
        except:  # noqa: E722
            log.exception('failed to get info for voevent-broker-peers')
            peers = []
        return {'voevent-broker-peers': peers}


class Receiver(TwistedService):
    """VOEvent receiver.

    Run a Comet-based VOEvent receiver
    (:class:`comet.protocol.subscriber.VOEventSubscriberFactory`). Starts after
    the :class:`~gwcelery.voevent.bootsteps.Reactor` bootstep.

    A few :doc:`configuration options <configuration>` are available:

    * ``voevent_receiver_address``: The address to connect to, in
      :samp:`{host}:{port}` format.

    The list of active connections is made available :ref:`inspection
    <celery:worker-inspect>` with the ``gwcelery inspect stats`` command under
    the ``voevent-receiver-peers`` key.
    """

    name = 'VOEvent receiver'

    requires = TwistedService.requires + (
        'celery.worker.consumer.tasks:Tasks',)

    def create_service(self, consumer):
        from comet.icomet import IHandler
        from twisted.application.internet import TCPClient
        from zope.interface import implementer
        from .subscriber import VOEventSubscriberFactory

        @implementer(IHandler)
        class Handler:

            def __call__(self, event):
                from twisted.internet import reactor
                reactor.callInThread(
                    voevent_received.send, sender=None, xml_document=event)

        conf = consumer.app.conf
        local_ivo = get_local_ivo(consumer.app)
        host, port = get_host_port(conf['voevent_receiver_address'])
        self._factory = factory = VOEventSubscriberFactory(
            local_ivo=local_ivo, handlers=[Handler()])
        return TCPClient(host, port, factory)

    def info(self, consumer):
        try:
            peers = [
                b.transport.getPeer().host for b in self._factory.subscribers]
        except:  # noqa: E722
            log.exception('failed to get info for voevent-receiver-peers')
            peers = []
        return {'voevent-receiver-peers': peers}



"""IGWN alert client."""
import json

from celery.utils.log import get_task_logger

from ..igwn_alert.signals import igwn_alert_received
from .. import app
from .core import DispatchHandler
from . import gracedb

log = get_task_logger(__name__)


class _IGWNAlertDispatchHandler(DispatchHandler):
    def __call__(self, *keys, **kwargs):
        try:
            igwn_alert_topics = app.conf['igwn_alert_topics']
        except KeyError:
            igwn_alert_topics = app.conf['igwn_alert_topics'] = set()
        igwn_alert_topics.update(keys)
        return super().__call__(*keys, **kwargs)

    def process_args(self, topic, alert):
        alert = json.loads(alert)
        # Determine GraceDB service URL
        try:
            try:
                self_link = alert['object']['links']['self']
            except KeyError:
                self_link = alert['object']['self']
        except KeyError:
            log.exception(
                'IGWN alert message does not contain an API URL: %r',
                alert)
            return None, None, None
        base, api, _ = self_link.partition('/api/')
        service = base + api

        if service != gracedb.client.url:
            # FIXME: this is probably redundant since IGWN alert client
            # is initialized using group gracedb-playground, gracedb-test etc.
            log.warning(
                'ignoring IGWN alert message because it is intended for '
                'GraceDB server %s, but we are set up for server %s',
                service, gracedb.client.url)
            return None, None, None

        return super().process_args(topic, alert)


handler = _IGWNAlertDispatchHandler()
r"""Function decorator to register a handler callback for specified IGWN alert
message types. The decorated function is turned into a Celery task, which will
be automatically called whenever a matching IGWN alert message is received.

Parameters
----------
\*keys
    List of IGWN alert message types to accept
\*\*kwargs
    Additional keyword arguments for :meth:`celery.Celery.task`.

Examples
--------
Declare a new handler like this::

    @igwn_alert.handler('cbc_gstlal',
                        'cbc_spiir',
                        'cbc_pycbc',
                        'cbc_mbta')
    def handle_cbc(alert_content):
        # do work here...
"""


@igwn_alert_received.connect
def _on_igwn_received(topic, payload, **kwargs):
    handler.dispatch(topic, payload)



import logging
import os
import stat
from unittest.mock import patch
from urllib.parse import urlparse, urlunparse

import pytest

from .. import __version__, app, sentry


@pytest.fixture
def netrcfile(monkeypatch, tmpdir):
    filename = str(tmpdir / 'netrc')
    monkeypatch.setenv('NETRC', filename)
    with open(filename, 'w') as f:
        os.fchmod(f.fileno(), stat.S_IRWXU)
    yield filename


@patch('sentry_sdk.init')
@patch('gwcelery.sentry.integrations.condor.CondorIntegration')
@patch('gwcelery.sentry.integrations.requests.RequestsIntegration')
@patch('gwcelery.sentry.integrations.subprocess.SubprocessIntegration')
@patch('sentry_sdk.integrations.celery.CeleryIntegration')
@patch('sentry_sdk.integrations.flask.FlaskIntegration')
@patch('sentry_sdk.integrations.redis.RedisIntegration')
@patch('sentry_sdk.integrations.tornado.TornadoIntegration')
def test_sentry_configure(mock_tornado_integration, mock_redis_integration,
                          mock_flask_integration, mock_celery_integration,
                          mock_subprocess_integration,
                          mock_requests_integration, mock_condor_integration,
                          mock_sdk_init, netrcfile, caplog):
    caplog.set_level(logging.ERROR)
    sentry.configure()
    record, = caplog.records
    assert 'Disabling Sentry integration' in record.message
    mock_celery_integration.assert_not_called()
    mock_sdk_init.assert_not_called()

    scheme, netloc, *rest = urlparse(sentry.DSN)
    with open(netrcfile, 'w') as f:
        print('machine', netloc, 'login', 'foo', 'password', 'bar', file=f)
    sentry.configure()
    dsn = urlunparse((scheme, 'foo@' + netloc, *rest))
    environment = app.conf['sentry_environment']
    release = 'gwcelery-' + __version__
    mock_celery_integration.assert_called_once_with()
    mock_sdk_init.assert_called_once_with(
        dsn, environment=environment, release=release,
        integrations=[
            mock_celery_integration.return_value,
            mock_condor_integration.return_value,
            mock_flask_integration.return_value,
            mock_redis_integration.return_value,
            mock_requests_integration.return_value,
            mock_subprocess_integration.return_value,
            mock_tornado_integration.return_value])



"""Create mock external events to be in coincidence
   with MDC superevents."""

from astropy.time import Time
from lxml import etree
import numpy as np
from pathlib import Path
import random
import re

from ..import app
from . import external_triggers
from . import igwn_alert


def create_external_event(gpstime, pipeline, se_search):
    """ Create a random external event VOEvent.

    Parameters
    ----------
    gpstime : float
        Event's gps time
    pipeline : str
        External trigger pipeline name
    se_search : str
        Search field for preferred event, 'MDC' or 'AllSky'
    """
    new_date = str(Time(gpstime, format='gps', scale='utc').isot) + 'Z'
    new_TrigID = str(int(gpstime))

    if pipeline in {'Fermi', 'Swift', 'INTEGRAL', 'AGILE'}:
        is_grb = True
    else:
        is_grb = False

    fname = str(Path(__file__).parent /
                '../tests/data/{0}_{1}gcn.xml'.format(
                    pipeline.lower(),
                    'grb_' if is_grb else ''))

    root = etree.parse(fname)
    # Change ivorn to indicate if this is an MDC event or O3 replay event
    root.xpath('.')[0].attrib['ivorn'] = \
        'ivo://lvk.internal/{0}#{1}_event{2}'.format(
            pipeline if pipeline != 'Swift' else 'SWIFT',
            'MDC-test' if se_search == 'MDC' else 'O3-replay',
            new_date).encode()

    # Change times to chosen time
    root.find("./Who/Date").text = str(new_date).encode()
    root.find(("./WhereWhen/ObsDataLocation/"
               "ObservationLocation/AstroCoords/Time/TimeInstant/"
               "ISOTime")).text = str(new_date).encode()
    root.find("./What/Param[@name='TrigID']").attrib['value'] = \
        str(new_TrigID).encode()

    if is_grb:
        # Give random sky position
        root.find(("./WhereWhen/ObsDataLocation/"
                   "ObservationLocation/AstroCoords/Position2D/Value2/"
                   "C1")).text = str(random.uniform(0, 360)).encode()
        thetas = np.arange(-np.pi / 2, np.pi / 2, .01)
        root.find(("./WhereWhen/ObsDataLocation/"
                   "ObservationLocation/AstroCoords/Position2D/Value2/"
                   "C2")).text = \
            str(random.choices(
                np.rad2deg(thetas),
                weights=np.cos(thetas) / sum(np.cos(thetas)))[0]).encode()
        if pipeline != 'Swift':
            root.find(
                ("./WhereWhen/ObsDataLocation/"
                 "ObservationLocation/AstroCoords/Position2D/"
                 "Error2Radius")).text = str(random.uniform(1, 30)).encode()

    return etree.tostring(root, xml_declaration=True, encoding="UTF-8",
                          pretty_print=True)


def _offset_time(gpstime, group):
    """ This function checks coincident time windows for superevents if they
    are of Burst or CBC group.

       Parameters
       ----------
       gpstime : float
           Event's gps time
       group : str
           Burst or CBC
    """
    if group == 'Burst':
        th, tl = app.conf['raven_coincidence_windows']['GRB_Burst']
    elif group == 'CBC':
        th, tl = app.conf['raven_coincidence_windows']['GRB_CBC']
    else:
        raise AssertionError(
            'Invalid group {}. Use only CBC or Burst.'.format(group))
    return gpstime + random.uniform(-tl, -th)


def _is_joint_mdc(graceid, se_search):
    """Upload external events to the user-defined frequency of MDC or AllSky
    superevents.

    Looks at the ending letters of a superevent (e.g. 'ac' from 'MS190124ac'),
    converts to a number, and checks if divisible by a number given in the
    configuration file.

    For example, if the configuration number
    :obj:`~gwcelery.conf.joint_mdc_freq` is 10,
    this means joint events with superevents ending with 'j', 't', 'ad', etc.
    """
    end_string = re.split(r'\d+', graceid)[-1].lower()
    val = 0
    for i in range(len(end_string)):
        val += (ord(end_string[i]) - 96) * 26 ** (len(end_string) - i - 1)
    return val % int(app.conf['joint_mdc_freq']) == 0 if se_search == 'MDC' \
        else val % int(app.conf['joint_O3_replay_freq']) == 0


@igwn_alert.handler('mdc_superevent',
                    'superevent',
                    queue='exttrig',
                    shared=False)
def upload_external_event(alert):
    """Upload a random GRB event for a certain percentage of MDC
    or O3-replay superevents.

    Notes
    -----
    Every n superevents, upload a GRB candidate within the
    standard CBC-GRB or Burst-GRB search window, where the frequency n is
    determined by the configuration variable
    :obj:`~gwcelery.conf.joint_mdc_freq` or
    :obj:`~gwcelery.conf.joint_O3_replay_freq`.

    For O3 replay testing with RAVEN pipeline, only run on gracedb-playground.
    """
    if alert['alert_type'] != 'new':
        return
    se_search = alert['object']['preferred_event_data']['search']
    group = alert['object']['preferred_event_data']['group']
    is_gracedb_playground = app.conf['gracedb_host'] \
        == 'gracedb-playground.ligo.org'
    joint_mdc_alert = se_search == 'MDC' and _is_joint_mdc(alert['uid'], 'MDC')
    joint_allsky_alert = se_search == 'AllSky' and \
        _is_joint_mdc(alert['uid'], 'AllSky') and is_gracedb_playground
    if not (joint_mdc_alert or joint_allsky_alert):
        return

    # Potentially upload 1, 2, or 3 GRB events
    num = 1 + np.random.choice(np.arange(3), p=[.6, .3, .1])
    events = []
    pipelines = []
    for i in range(num):
        gpstime = float(alert['object']['t_0'])
        new_time = _offset_time(gpstime, group)

        # Choose external grb pipeline to simulate
        pipeline = np.random.choice(['Fermi', 'Swift', 'INTEGRAL', 'AGILE'],
                                    p=[.5, .3, .1, .1])
        ext_event = create_external_event(new_time, pipeline, se_search)

        # Upload as from GCN
        external_triggers.handle_grb_gcn(ext_event)

        events.append(ext_event), pipelines.append(pipeline)

    return events, pipelines


@app.task(queue='exttrig',
          ignore_result=True,
          shared=False)
def upload_snews_event():
    """Create and upload a SNEWS-like MDC external event."""
    current_time = Time(Time.now(), format='gps').value
    ext_event = create_external_event(current_time, 'SNEWS', 'MDC')
    external_triggers.handle_snews_gcn(ext_event)
    return ext_event



from importlib import resources
from unittest.mock import patch

from astropy.table import Table
import numpy as np
import pytest

from . import data
from ..util import read_json
from .test_tasks_skymaps import toy_fits_filecontents  # noqa: F401
from .test_tasks_skymaps import toy_3d_fits_filecontents  # noqa: F401
from ..tasks import external_skymaps


true_heasarc_link = ('http://heasarc.gsfc.nasa.gov/FTP/fermi/data/gbm/'
                     + 'triggers/2017/bn170817529/current/')
true_skymap_link = true_heasarc_link + 'glg_healpix_all_bn170817529_v00.fit'


def mock_get_event(exttrig):
    return {'search': 'GRB'}


def mock_get_superevent(graceid):
    return read_json(data, 'mock_superevent_object.json')


def mock_get_log(graceid):
    if graceid == 'S12345':
        return read_json(data, 'gracedb_setrigger_log.json')
    elif graceid == 'E12345':
        return read_json(data, 'gracedb_externaltrigger_log.json')
    else:
        raise ValueError


@pytest.fixture  # noqa: F811
def mock_download(monkeypatch, toy_3d_fits_filecontents):  # noqa: F811

    def download(filename, graceid):
        """Mocks GraceDB download functionality"""
        if graceid == 'S12345' and filename == 'bayestar.fits.gz,0':
            return toy_3d_fits_filecontents
        elif (graceid == 'E12345' and
              filename == ('nasa.gsfc.gcn_Fermi%23GBM_Gnd_Pos_2017-08-17'
                           + 'T12%3A41%3A06.47_524666471_57-431.xml')):
            return resources.read_binary(
                data, 'externaltrigger_original_data.xml')
        else:
            raise ValueError

    monkeypatch.setattr('gwcelery.tasks.gracedb.download.run', download)


def mock_get_file_contents(monkeypatch, toy_fits_filecontents):  # noqa: F811
    """Mocks astropy get_file_contents functionality"""
    def get_file_contents(heasarc_link):
        assert heasarc_link == true_heasarc_link
        return toy_fits_filecontents

    monkeypatch.setattr(
        'astropy.utils.data.get_file_contents', get_file_contents)


def get_gw_moc_skymap():
    array = [np.arange(12, dtype=np.float64)] * 5
    #  Modify UNIQ table to be allowable values
    array[4] = array[4] + 4
    table = Table(
        array,
        names=['PROBDENSITY', 'DISTMU', 'DISTSIGMA', 'DISTNORM', 'UNIQ'])
    table.meta['comment'] = 'This is a comment.'
    table.meta['HISTORY'] = \
        ['This is a history line. <This should be escaped.>']
    table.meta['OBJECT'] = 'T12345'
    table.meta['LOGBCI'] = 3.5
    table.meta['ORDERING'] = 'NESTED'
    table.meta['instruments'] = {'L1', 'H1', 'V1'}
    return table


@patch('gwcelery.tasks.skymaps.plot_allsky.run')
@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.external_skymaps.combine_skymaps.run')
@patch('gwcelery.tasks.gracedb.download')
@patch('gwcelery.tasks.external_skymaps.get_skymap_filename',
       return_value='fermi_skymap.fits.gz,0')
def test_create_combined_skymap(mock_get_skymap_filename,
                                mock_download,
                                mock_combine_skymaps, mock_upload,
                                mock_plot_allsky):
    """Test creating combined LVC and Fermi skymap"""
    # Run function under test
    external_skymaps.create_combined_skymap('S12345', 'E12345')
    mock_combine_skymaps.assert_called_once()
    mock_upload.assert_called()


def _mock_read_sky_map(filename, moc=True):
    if moc:
        return get_gw_moc_skymap()
    else:
        ext_sky = np.full(12, 1 / 12)
        ext_header = {'instruments': set({'Fermi'}), 'nest': True}
        return ext_sky, ext_header


@pytest.mark.parametrize('gw_moc',
                         [True, False])
@patch('ligo.skymap.tool.ligo_skymap_combine.main')
@patch('gwcelery.tasks.external_skymaps.combine_skymaps_moc_flat')
@patch('ligo.skymap.io.fits.read_sky_map', side_effect=_mock_read_sky_map)
@patch('ligo.skymap.io.fits.write_sky_map')
def test_combine_skymaps(mock_write_sky_map,
                         mock_read_sky_map,
                         mock_skymap_combine_moc_flat,
                         mock_skymap_combine_flat_flat,
                         gw_moc):
    """Test using our internal MOC-flat sky map combination gives back the
    input using a uniform sky map, ensuring the test is giving a sane result
    and is at least running to completion.
    """
    external_skymaps.combine_skymaps((b'', b''), gw_moc=gw_moc)
    if gw_moc:
        mock_read_sky_map.assert_called()
        mock_skymap_combine_moc_flat.assert_called_once()
        mock_write_sky_map.assert_called_once()
    else:
        mock_skymap_combine_flat_flat.assert_called()


@pytest.mark.parametrize('missing_header_values,instrument',
                         [[False, 'Fermi'],
                          [True, 'Fermi'],
                          [False, None],
                          [True, None]])
def test_create_combined_skymap_moc_flat(missing_header_values, instrument):
    """Test using our internal MOC-flat sky map combination gives back the
    input using a uniform sky map, ensuring the test is giving a sane result
    and is at least running to completion.
    """
    # Run function under test
    gw_sky = get_gw_moc_skymap()
    if missing_header_values:
        del gw_sky['DISTMU']
        del gw_sky['DISTSIGMA']
        gw_sky.meta.pop('instruments')
        gw_sky.meta.pop('HISTORY')
    ext_sky = np.full(12, 1 / 12)
    if instrument:
        ext_header = {'instruments': set({instrument}), 'nest': True}
    else:
        ext_header = {'nest': True}
    combined_sky = external_skymaps.combine_skymaps_moc_flat(gw_sky, ext_sky,
                                                             ext_header)
    assert all(combined_sky['PROBDENSITY'] == gw_sky['PROBDENSITY'])
    if missing_header_values:
        assert 'instruments' not in combined_sky.meta
    else:
        assert ('Fermi' in combined_sky.meta['instruments'] if instrument else
                'external instrument' in combined_sky.meta['instruments'])


@pytest.mark.parametrize('graceid',
                         ['S12345', 'E12345'])
@patch('gwcelery.tasks.gracedb.get_log', side_effect=mock_get_log)
def test_get_skymap_filename(mock_get_logs, graceid):
    """Test getting the LVC skymap fits filename"""
    filename = external_skymaps.get_skymap_filename(graceid,
                                                    is_gw='S' in graceid)
    if 'S' in graceid:
        assert filename == 'bayestar.multiorder.fits,0'
    elif 'E' in graceid:
        assert filename == 'fermi_skymap.fits.gz,0'


@patch('gwcelery.tasks.gracedb.get_event', mock_get_event)
@patch('gwcelery.tasks.gracedb.get_superevent',
       return_value={'em_events': ['E12345']})
def test_external_trigger(mock_get_superevent, mock_download):
    """Test getting related em event for superevent"""
    assert external_skymaps.external_trigger('S12345') == 'E12345'


@patch('gwcelery.tasks.gracedb.get_log', mock_get_log)
def test_external_trigger_heasarc(mock_download):
    """Test retrieving HEASARC fits file link from GCN"""
    heasarc_link = external_skymaps.external_trigger_heasarc('E12345')
    assert heasarc_link == true_heasarc_link


@pytest.mark.parametrize('search', ['GRB', 'SubGRB', 'FromURL'])
@patch('urllib.request.urlopen')
def test_get_external_skymap(mock_urlopen, search):
    """Assert that the correct call to astropy.get_file_contents is used"""
    external_skymaps.get_external_skymap(true_heasarc_link, search)
    mock_urlopen.assert_called_once()


@pytest.mark.parametrize('search', ['GRB', 'SubGRB', 'FromURL'])
@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.skymaps.plot_allsky.run')
@patch('gwcelery.tasks.external_skymaps.get_external_skymap.run')
@patch('gwcelery.tasks.external_skymaps.external_trigger_heasarc.run')
def test_get_upload_external_skymap(mock_external_trigger_heasarc,
                                    mock_get_external_skymap,
                                    mock_plot_allsky,
                                    mock_upload,
                                    search):
    """Test that an external sky map is grabbed and uploaded."""
    event = {'graceid': 'E12345', 'search': search}
    external_skymaps.get_upload_external_skymap(event)
    if search == 'GRB':
        mock_external_trigger_heasarc.assert_called_once()
    mock_get_external_skymap.assert_called_once()
    mock_upload.assert_called()


@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.skymaps.plot_allsky.run')
@patch('gwcelery.tasks.external_skymaps.get_external_skymap.run')
def test_get_upload_external_skymap_subgrb(mock_get_external_skymap,
                                           mock_plot_allsky,
                                           mock_upload):
    """Test that an external sky map is grabbed and uploaded."""
    event = {'graceid': 'E12345', 'search': 'SubGRB'}
    external_skymaps.get_upload_external_skymap(
        event,
        ('https://gcn.gsfc.nasa.gov/notices_gbm_sub/' +
         'gbm_subthresh_604671025.728000_healpix.fits'))
    mock_get_external_skymap.assert_called_once()
    mock_upload.assert_called()


@pytest.mark.parametrize('ra,dec,error,pix',
                         [[0, 90, 0, 0],
                          [270, -90, .01, -1]])
def test_create_swift_skymap(ra, dec, error, pix):
    """Test created single pixel sky maps for Swift localization."""
    skymap = external_skymaps.create_external_skymap(ra, dec, error, 'Swift')
    assert skymap[pix] == 1


def test_create_fermi_skymap():
    """Test created single pixel sky maps for Swift localization."""
    ra, dec, error = 0, 90, 10
    assert (np.sum(external_skymaps.create_external_skymap(
               ra, dec, error, 'Fermi')) ==
           pytest.approx(1.0, 1.e-9))


@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.skymaps.plot_allsky.run')
def test_create_upload_swift_skymap(mock_plot_allsky,
                                    mock_upload):
    """Test the creation and upload of sky maps for Swift localization."""
    event = {'graceid': 'E1234',
             'pipeline': 'Swift',
             'gpstime': 1259790538.77,
             'extra_attributes': {
                 'GRB': {
                     'trigger_id': 1234567,
                     'ra': 1.,
                     'dec': 1.,
                     'error_radius': 0}},
             'links': {
                 'self': 'https://gracedb.ligo.org/api/events/E356793'}}
    external_skymaps.create_upload_external_skymap(event, '111',
                                                   '2020-01-09T01:47:09')
    mock_upload.assert_called()
    mock_plot_allsky.assert_called_once()


@patch('gwcelery.tasks.gracedb.upload.run')
def test_create_upload_skymap_filter(mock_upload):
    """Test that empty notices don't create sky maps."""
    event = {'graceid': 'E1234',
             'pipeline': 'Swift',
             'gpstime': 1259790538.77,
             'extra_attributes': {
                 'GRB': {
                     'trigger_id': 1234567,
                     'ra': 0.,
                     'dec': 0.,
                     'error_radius': 0.}},
             'links': {
                 'self': 'https://gracedb.ligo.org/api/events/E356793'}}
    external_skymaps.create_upload_external_skymap(event, '111',
                                                   '2020-01-09T01:47:09')
    mock_upload.assert_not_called()


@pytest.mark.parametrize(
    'em_type,graceid,labels,expected_result',
    [[None, 'E1', [], True],
     ['E1', 'E1', [], True],
     ['E1', 'E2', [], True],
     ['E1', 'E1', ['RAVEN_ALERT'], True],
     ['E1', 'E2', ['RAVEN_ALERT'], False],
     [None, 'E3', [], False]]
)
@patch('gwcelery.tasks.gracedb.upload.run')
def test_plot_overlap_integral(mock_upload,
                               em_type, graceid, labels, expected_result):

    coinc_far_dict = {'skymap_overlap': 1e2} if graceid != 'E3' else {}
    superevent = {'superevent_id': 'S1', 'em_type': em_type, 'labels': labels}
    ext_event = {'graceid': graceid}
    external_skymaps.plot_overlap_integral(coinc_far_dict, superevent,
                                           ext_event)
    if expected_result:
        mock_upload.assert_called_once()
    else:
        mock_upload.assert_not_called()
